import os
from awscli.clidriver import create_clidriver
from typing import Mapping


# https://github.com/boto/boto3/issues/358#issuecomment-372086466
def aws_cli(*cmd: str) -> None:
    old_env: Mapping[str, str] = dict(os.environ)
    try:
        # Environment
        env = os.environ.copy()
        env['LC_CTYPE'] = u'en_US.UTF'
        os.environ.update(env)

        # Run awscli in the same process
        exit_code = create_clidriver().main(args=cmd)

        # Deal with problems
        if exit_code > 0:
            raise RuntimeError('AWS CLI exited with code {}'.format(exit_code))
    finally:
        os.environ.clear()
        os.environ.update(old_env)



from contextlib import contextmanager
from records_mover.records.prep import TablePrep, TargetTableDetails
from records_mover.records.records_directory import RecordsDirectory
from records_mover.records.processing_instructions import ProcessingInstructions
from records_mover.records.results import MoveResult
from records_mover.url.base import BaseDirectoryUrl
from records_mover.records.sources import SupportsMoveToRecordsDirectory
from records_mover.records.targets.table.base import BaseTableMoveAlgorithm
import logging
from typing import Iterator, TYPE_CHECKING
if TYPE_CHECKING:
    from .target import TableRecordsTarget  # Dodge circular dependency

logger = logging.getLogger(__name__)


class DoMoveFromTempLocAfterFillingIt(BaseTableMoveAlgorithm):
    def __init__(self,
                 prep: TablePrep,
                 target_table_details: TargetTableDetails,
                 table_target: 'TableRecordsTarget',
                 records_source: SupportsMoveToRecordsDirectory,
                 processing_instructions: ProcessingInstructions) -> None:
        self.table_target = table_target
        self.records_source = records_source
        super().__init__(prep, target_table_details, processing_instructions)

    @contextmanager
    def temporary_loadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        driver = self.tbl.db_driver(self.tbl.db_engine)
        loader = driver.loader()
        # This will only be reached in move() if
        # Source#has_compatible_format(records_target) returns true,
        # which means we were able to get a loader and call
        # can_load_this_format() previously.
        assert loader is not None
        if loader.has_temporary_loadable_directory_loc():
            with loader.temporary_loadable_directory_loc() as loc:
                yield loc
        else:
            # As a last resort, use the source's temporary directory
            with self.records_source.temporary_unloadable_directory_loc() as loc:
                yield loc

    def move(self) -> MoveResult:
        pis = self.processing_instructions
        records_format = self.records_source.compatible_format(self.table_target)
        if records_format is None:
            raise NotImplementedError("No compatible records format between "
                                      f"{self.records_source} and {self}")
        with self.temporary_loadable_directory_loc() as temp_loc:
            directory = RecordsDirectory(records_loc=temp_loc)
            self.records_source.\
                move_to_records_directory(records_directory=directory,
                                          records_format=records_format,
                                          processing_instructions=pis)
            out = self.table_target.\
                move_from_records_directory(directory=directory,
                                            processing_instructions=self.processing_instructions)
            return out



import logging
from sqlalchemy.schema import CreateTable, MetaData
from sqlalchemy import Table
from typing import Dict, TYPE_CHECKING
from .known_representation import RecordsSchemaKnownRepresentation
from ..field import RecordsSchemaField
if TYPE_CHECKING:
    from ....db import DBDriver  # noqa
    from ..schema import RecordsSchema  # noqa


logger = logging.getLogger(__name__)


def schema_to_schema_sql(records_schema: 'RecordsSchema',
                         driver: 'DBDriver',
                         schema_name: str,
                         table_name: str) -> str:
    meta = MetaData()
    columns = [f.to_sqlalchemy_column(driver) for f in records_schema.fields]
    table = Table(table_name, meta,
                  *columns,
                  schema=schema_name)
    db = None
    if driver is not None:
        db = driver.db
    return str(CreateTable(table, bind=db))


def schema_from_db_table(schema_name: str,
                         table_name: str,
                         driver: 'DBDriver') -> 'RecordsSchema':
    from ..schema import RecordsSchema  # noqa
    logger.info('Pulling table metadata...')
    table = driver.table(schema_name, table_name)
    logger.info(f"Table metadata: {table}")

    fields = []
    origin_representation = RecordsSchemaKnownRepresentation.from_db_driver(driver,
                                                                            schema_name,
                                                                            table_name)
    known_representations: Dict[str, RecordsSchemaKnownRepresentation] = {
        'origin': origin_representation
    }

    for column in table.columns:
        fields.append(RecordsSchemaField.
                      from_sqlalchemy_column(column=column,
                                             driver=driver,
                                             rep_type=origin_representation.type))

    return RecordsSchema(fields=fields,
                         known_representations=known_representations)



import warnings


def warn_deprecated(**kwargs) -> None:
    for name, value in kwargs.items():
        if value is not None:
            warnings.simplefilter('always', DeprecationWarning)  # turn off filter
            warnings.warn(f"{name} argument is deprecated, will not be used, and "
                          "will be removed in a future release",
                          category=DeprecationWarning,
                          stacklevel=4)
            warnings.simplefilter('default', DeprecationWarning)  # reset filter



from typing import NamedTuple, TypeVar
from .types import (UntypedRecordsHints, HintHeaderRow, HintFieldDelimiter,
                    HintCompression, HintRecordTerminator,
                    HintQuoting, HintQuoteChar, HintDoublequote,
                    HintEscape, HintEncoding, HintDateFormat,
                    HintTimeOnlyFormat, HintDateTimeFormatTz,
                    HintDateTimeFormat)
from .hints import Hints
from .hint import Hint


class ValidatedRecordsHints(NamedTuple):
    header_row: HintHeaderRow
    field_delimiter: HintFieldDelimiter
    compression: HintCompression
    record_terminator: HintRecordTerminator
    quoting: HintQuoting
    quotechar: HintQuoteChar
    doublequote: HintDoublequote
    escape: HintEscape
    encoding: HintEncoding
    dateformat: HintDateFormat
    timeonlyformat: HintTimeOnlyFormat
    datetimeformattz: HintDateTimeFormatTz
    datetimeformat: HintDateTimeFormat

    @staticmethod
    def validate(hints: UntypedRecordsHints,
                 fail_if_cant_handle_hint: bool) -> 'ValidatedRecordsHints':
        T = TypeVar('T')

        def v(hint: Hint[T]) -> T:
            return hint.validate(hints, fail_if_cant_handle_hint)

        return ValidatedRecordsHints(
            header_row=v(Hints.header_row.value),
            field_delimiter=v(Hints.field_delimiter.value),
            compression=v(Hints.compression.value),
            record_terminator=v(Hints.record_terminator.value),
            quoting=v(Hints.quoting.value),
            quotechar=v(Hints.quotechar.value),
            doublequote=v(Hints.doublequote.value),
            escape=v(Hints.escape.value),
            encoding=v(Hints.encoding.value),
            dateformat=v(Hints.dateformat.value),
            timeonlyformat=v(Hints.timeonlyformat.value),
            datetimeformattz=v(Hints.datetimeformattz.value),
            datetimeformat=v(Hints.datetimeformat.value),
        )



import pathlib
from ...url.resolver import UrlResolver
from ..records_format import BaseRecordsFormat
from .fileobj import FileobjTarget
from .directory_from_url import DirectoryFromUrlRecordsTarget
from .data_url import DataUrlTarget
from typing import Callable, Optional, Union, Dict, List, IO, TYPE_CHECKING
from ..existing_table_handling import ExistingTableHandling
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa
    from sqlalchemy.engine import Engine, Connection  # noqa
    from ...db import DBDriver  # noqa
    from .spectrum import SpectrumRecordsTarget  # noqa
    from .google_sheets import GoogleSheetsRecordsTarget  # noqa
    from .table import TableRecordsTarget  # noqa
    from .directory_from_url import DirectoryFromUrlRecordsTarget  # noqa


class RecordsTargets(object):
    """These methods produce objects representing the target of a records
    move.  The objects can be used as the 'target' argument to
    :meth:`records_mover.records.move`

    This object should be pulled from the 'targets' property of the
    'records' property on a :class:`records_mover.Session` object
    instead of being constructed directly.

    Example use:

    .. code-block:: python

       records = session.records
       db_engine = session.get_default_db_engine()
       url = 's3://some-bucket/some-directory/'
       source = records.sources.directory_from_url(url=url)
       target = records.targets.table(schema_name='myschema',
                                      table_name='mytable',
                                      db_engine=db_engine)
       results = records.move(source, target)

    """
    def __init__(self,
                 url_resolver: UrlResolver,
                 db_driver: Callable[[Union['Engine', 'Connection']], 'DBDriver']) -> None:
        self.url_resolver = url_resolver
        self.db_driver = db_driver

    def directory_from_url(self,
                           output_url: str,
                           records_format:
                           Optional[BaseRecordsFormat]=None) ->\
            'DirectoryFromUrlRecordsTarget':
        """Represents a Records Directory pointed to by a URL as a target.

        :param output_url: Location to write the records directory.  Must be a URL format
           understood by the records_mover.url library, and must be a directory URL that ends with
           a '/'.
        :param records_format: Description of the format of the data files to write out.  If not
           specified, an efficient format for bulk moves will be chosen.
        """
        from .directory_from_url import DirectoryFromUrlRecordsTarget  # noqa
        return DirectoryFromUrlRecordsTarget(output_url=output_url,
                                             records_format=records_format,
                                             url_resolver=self.url_resolver)

    def table(self,
              db_engine: 'Engine',
              schema_name: str,
              table_name: str,
              existing_table_handling: ExistingTableHandling=
              ExistingTableHandling.DELETE_AND_OVERWRITE,
              drop_and_recreate_on_load_error: bool=False,
              add_user_perms_for: Optional[Dict[str, List[str]]]=None,
              add_group_perms_for: Optional[Dict[str, List[str]]]=None) -> \
            'TableRecordsTarget':
        """Represents a SQLALchemy-accessible database table as as a target.

        :param db_engine: SQLAlchemy database engine to write data to.

        :param schema_name: Schema name of a table to write data to.

        :param table_name: Table name of a table to write data to.

        :param existing_table_handling: When loading into a database table, controls how any
           existing table found will be handled.  This must be a
           :class:`records_mover.records.ExistingTableHandling` object.

        :param drop_and_recreate_on_load_error: If True, table load errors will attempt to be
           addressed by dropping the target table and reloading the incoming data.

        :param add_user_perms_for: If specified, a table's permissions will be set for the
           specified users. Format should be like {'all': ['username1', 'username2'], 'select':
           ['username3', 'username4']}

        :param add_group_perms_for: If specified, a table's permissions will be set for the
           specified group. Format should be like {'all': ['group1', 'group2'], 'select': ['group3',
           'group4']}
        """
        from .table import TableRecordsTarget  # noqa
        return TableRecordsTarget(schema_name=schema_name,
                                  table_name=table_name,
                                  db_engine=db_engine,
                                  db_driver=self.db_driver,
                                  existing_table_handling=existing_table_handling,
                                  drop_and_recreate_on_load_error=drop_and_recreate_on_load_error,
                                  add_user_perms_for=add_user_perms_for,
                                  add_group_perms_for=add_group_perms_for)

    def google_sheet(self,
                     spreadsheet_id: str,
                     sheet_name: str,
                     google_cloud_creds:
                     'google.auth.credentials.Credentials') ->\
            'GoogleSheetsRecordsTarget':
        """Represents a sheet in a Google Sheets spreadsheet as a target, via
        the Google Sheets API.

        :param spreadsheet_id: This is the xyz in
           https://docs.google.com/spreadsheets/d/xyz/edit?ts=5be5b383#gid=abc
        :param sheet_name: This is the label of the particular tab within the Google Sheets
           spreadsheet where the data should go.
        :param google_cloud_creds: Credentials object for Google Cloud Platform access.

        """
        # see the 'gsheets' extras_require option in setup.py - needed for this!
        from .google_sheets import GoogleSheetsRecordsTarget  # noqa
        return GoogleSheetsRecordsTarget(spreadsheet_id=spreadsheet_id,
                                         sheet_name=sheet_name,
                                         google_cloud_creds=google_cloud_creds)

    def fileobj(self,
                output_fileobj: IO[bytes],
                records_format: BaseRecordsFormat) -> FileobjTarget:
        """Represents a stream of data files bytes as a target.

        :param output_fileobj: Stream where the file shoud be written to.
        :param records_format: Description of the format of the data files to write out.  If not
           specified, an efficient format for bulk moves will be chosen.
        """
        return FileobjTarget(fileobj=output_fileobj, records_format=records_format)

    def data_url(self,
                 output_url: str,
                 records_format: Optional[BaseRecordsFormat]=None) -> DataUrlTarget:
        """Represents a URL pointer to a data file as a target.

        :param output_url: Location of the data file to write.  Must be a URL format understood by
           the records_mover.url library corresponding to a file, not a directory (i.e., not ending
           with a '/')
        :param records_format: Description of the format of the data files to write out.  If not
           specified, an efficient format for bulk moves will be chosen.
        """
        output_loc = self.url_resolver.file_url(output_url)
        return DataUrlTarget(output_loc=output_loc,
                             records_format=records_format)

    def local_file(self,
                   filename: str,
                   records_format: Optional[BaseRecordsFormat]=None) ->\
            'DataUrlTarget':
        """Represents a data file on the local filesystem as a target.

        :param filename: File path (relative or absolute) of the data file to unload to.
        :param records_format: Description of the format of the data files to write out.  If not
           specified, an efficient format for bulk moves will be chosen.
        """
        url = pathlib.Path(filename).resolve().as_uri()
        return self.data_url(output_url=url, records_format=records_format)

    def spectrum(self,
                 schema_name: str,
                 table_name: str,
                 db_engine: 'Engine',
                 spectrum_base_url: Optional[str]=None,
                 spectrum_rdir_url: Optional[str]=None,
                 existing_table_handling: ExistingTableHandling=
                 ExistingTableHandling.TRUNCATE_AND_OVERWRITE) ->\
            'SpectrumRecordsTarget':
        """
        Represents a location in Amazon Redshift Spectrum as a target.

        :param schema_name: Schema name of a table to write data to.

        :param table_name: Table name of a table to write data to.

        :param db_engine: SQLAlchemy database engine to write data to.

        :param spectrum_base_url: Root S3 URL under which a simple directory structure will be
           created for files to be stored, if spectrum_rdir_url is not specified.  Note that when
           using the mover CLI, db-facts may be used to provide a default.

        :param spectrum_rdir_url: S3 URL where a records directory with files will be stored;
           otherwise, use db-facts default if exists.  If this is not specified, spectrum_base_url
           must be.

        :param existing_table_handling: When loading into a database table, controls how any
           existing table found will be handled.  This must be a
           :class:`records_mover.records.ExistingTableHandling` object.
        """
        from .spectrum import SpectrumRecordsTarget  # noqa

        return SpectrumRecordsTarget(schema_name=schema_name,
                                     table_name=table_name,
                                     db_engine=db_engine,
                                     db_driver=self.db_driver,
                                     url_resolver=self.url_resolver,
                                     spectrum_base_url=spectrum_base_url,
                                     spectrum_rdir_url=spectrum_rdir_url,
                                     existing_table_handling=existing_table_handling)



from ..results import MoveResult
from ..records_directory import RecordsDirectory
from .base import (SupportsMoveFromDataframes,
                   SupportsMoveFromRecordsDirectory)
from ..processing_instructions import ProcessingInstructions
from ...url.base import BaseFileUrl
from ..records_format import (
    BaseRecordsFormat, DelimitedRecordsFormat, ParquetRecordsFormat, AvroRecordsFormat
)
from ..delimited import sniff_compression_from_url
from .fileobj import FileobjTarget
from typing import Optional, List, TYPE_CHECKING
if TYPE_CHECKING:
    from pandas import DataFrame  # noqa
    from ..sources.dataframes import DataframesRecordsSource


def sniff_records_format_from_url(url: str) -> BaseRecordsFormat:
    if url.endswith('.parquet'):
        return ParquetRecordsFormat()
    elif url.endswith('.avro'):
        return AvroRecordsFormat()
    else:
        return DelimitedRecordsFormat()


class DataUrlTarget(SupportsMoveFromDataframes,
                    SupportsMoveFromRecordsDirectory):
    def __init__(self,
                 output_loc: BaseFileUrl,
                 records_format: Optional[BaseRecordsFormat]) -> None:
        if records_format is None:
            records_format = sniff_records_format_from_url(output_loc.url)

        if (isinstance(records_format, DelimitedRecordsFormat) and
           'compression' not in records_format.custom_hints):
            # if user hasn't very specifically specified compression
            # type, don't surprise the user by writing a compression
            # they're not expecting.
            compression = sniff_compression_from_url(output_loc.url)
            inferred_hints = {
                'compression': compression
            }
            records_format = records_format.alter_hints(inferred_hints)

        self.records_format = records_format
        self.output_loc = output_loc

    def move_from_dataframes_source(self,
                                    dfs_source: 'DataframesRecordsSource',
                                    processing_instructions:
                                    ProcessingInstructions) -> MoveResult:
        with self.output_loc.open(mode='wb') as fileobj:
            fileobj_target = FileobjTarget(fileobj=fileobj, records_format=self.records_format)
            return fileobj_target.\
                move_from_dataframes_source(dfs_source=dfs_source,
                                            processing_instructions=processing_instructions)

    def can_move_directly_from_scheme(self, scheme: str) -> bool:
        # Currently all means of copying between different schemes
        # involve streaming data down to Records Mover from the source
        # and then back up to the target.
        return False

    def move_from_records_directory(self,
                                    directory: RecordsDirectory,
                                    processing_instructions: ProcessingInstructions,
                                    override_records_format: Optional[BaseRecordsFormat] = None)\
            -> MoveResult:
        records_format = directory.load_format(processing_instructions.fail_if_dont_understand)
        directory.save_to_url(self.output_loc)
        return MoveResult(move_count=None,
                          output_urls={
                              records_format.generate_filename('data'): self.output_loc.url
                          })

    def can_move_from_format(self,
                             source_records_format: BaseRecordsFormat) -> bool:
        if self.records_format is None:
            return True
        else:
            return self.records_format == source_records_format

    def known_supported_records_formats(self) -> List[BaseRecordsFormat]:
        if self.records_format is not None:
            return [self.records_format]
        return []



from typing import Dict, List
from typing_extensions import Literal, TypedDict
from typing_inspect import get_args


# https://docs.aws.amazon.com/redshift/latest/dg/r_CREATE_EXTERNAL_TABLE.html
class RecordsManifestEntryMetadata(TypedDict):
    content_length: int


class BaseRecordsManifestEntry(TypedDict):
    url: str
    mandatory: bool


class LegacyRecordsManifestEntry(BaseRecordsManifestEntry, total=False):
    meta: RecordsManifestEntryMetadata


class RecordsManifestEntryWithLength(BaseRecordsManifestEntry):
    meta: RecordsManifestEntryMetadata


class LegacyRecordsManifest(TypedDict):
    entries: List[LegacyRecordsManifestEntry]


class RecordsManifestWithLength(TypedDict):
    entries: List[RecordsManifestEntryWithLength]


class UrlDetailsEntry(TypedDict):
    content_length: int


Url = str
UrlDetails = Dict[Url, UrlDetailsEntry]


RecordsFormatType = Literal['avro', 'delimited', 'parquet']

RECORDS_FORMAT_TYPES: List[str] = list(get_args(RecordsFormatType))

DelimitedVariant = Literal['dumb', 'csv', 'bigquery', 'bluelabs', 'vertica']



from .base import SupportsRecordsDirectory
from typing import Optional
from ..records_directory import RecordsDirectory
from ...url.resolver import UrlResolver
from ..records_format import BaseRecordsFormat
from ..errors import RecordsFolderNonEmptyException


class DirectoryFromUrlRecordsTarget(SupportsRecordsDirectory):
    def __init__(self,
                 output_url: str,
                 url_resolver: UrlResolver,
                 records_format: Optional[BaseRecordsFormat]) -> None:
        if not output_url.endswith('/'):
            raise ValueError("Please provide a directory name - "
                             f"URL should end with '/': {output_url}")
        self.output_loc = url_resolver.directory_url(output_url)
        self.records_format = records_format

    def validate(self) -> None:
        if (len(self.output_loc.files_in_directory()) > 0):
            message = f"Target directory {self.output_loc.url} is non-empty."
            raise RecordsFolderNonEmptyException(message)

    def records_directory(self) -> RecordsDirectory:
        return RecordsDirectory(records_loc=self.output_loc)



from ...utils import quiet_remove
from ...records.delimited import cant_handle_hint
from typing import Dict, Any, Set
from sqlalchemy_redshift.commands import Format
from records_mover.mover_types import _assert_never
from ...records.records_format import (
    DelimitedRecordsFormat, ParquetRecordsFormat, BaseRecordsFormat
)

RedshiftUnloadOptions = Dict[str, Any]


# https://docs.aws.amazon.com/redshift/latest/dg/r_UNLOAD.html
#
def redshift_unload_options(unhandled_hints: Set[str],
                            records_format: BaseRecordsFormat,
                            fail_if_cant_handle_hint: bool) -> RedshiftUnloadOptions:
    redshift_options: RedshiftUnloadOptions = {}
    if isinstance(records_format, ParquetRecordsFormat):
        redshift_options['format'] = Format.parquet
        return redshift_options

    if not isinstance(records_format, DelimitedRecordsFormat):
        raise NotImplementedError("Redshift export only supported via Parquet and "
                                  "delimited currently")
    hints = records_format.\
        validate(fail_if_cant_handle_hint=fail_if_cant_handle_hint)
    if hints.escape == '\\':
        redshift_options['escape'] = True
    elif hints.escape is None:
        pass
    else:
        _assert_never(hints.escape)
    quiet_remove(unhandled_hints, 'escape')
    redshift_options['delimiter'] = hints.field_delimiter
    quiet_remove(unhandled_hints, 'field-delimiter')
    if hints.record_terminator == "\n":
        # This is Redshift's one and only export format
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'record-terminator', hints)
    quiet_remove(unhandled_hints, 'record-terminator')
    if hints.quoting == 'all':
        if hints.doublequote is not False:
            cant_handle_hint(fail_if_cant_handle_hint, 'doublequote', hints)
        if hints.quotechar != '"':
            cant_handle_hint(fail_if_cant_handle_hint, 'quotechar', hints)
        redshift_options['add_quotes'] = True
    elif hints.quoting is None:
        redshift_options['add_quotes'] = False
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
    quiet_remove(unhandled_hints, 'quoting')
    quiet_remove(unhandled_hints, 'doublequote')
    quiet_remove(unhandled_hints, 'quotechar')
    if hints.compression == 'GZIP':
        redshift_options['gzip'] = True
    elif hints.compression is None:
        # good to go
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)
    quiet_remove(unhandled_hints, 'compression')
    if hints.encoding != 'UTF8':
        cant_handle_hint(fail_if_cant_handle_hint, 'encoding', hints)
    quiet_remove(unhandled_hints, 'encoding')
    if hints.datetimeformattz not in ['YYYY-MM-DD HH:MI:SSOF',
                                      'YYYY-MM-DD HH24:MI:SSOF']:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformattz', hints)
    quiet_remove(unhandled_hints, 'datetimeformattz')
    if hints.datetimeformat not in ['YYYY-MM-DD HH24:MI:SS',
                                    'YYYY-MM-DD HH:MI:SS']:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', hints)
    quiet_remove(unhandled_hints, 'datetimeformat')
    if hints.dateformat != 'YYYY-MM-DD':
        cant_handle_hint(fail_if_cant_handle_hint, 'dateformat', hints)
    quiet_remove(unhandled_hints, 'dateformat')
    # Redshift doesn't have a time without date type, so there's
    # nothing that could be exported there.
    quiet_remove(unhandled_hints, 'timeonlyformat')

    if hints.header_row:
        # Redshift unload doesn't know how to add headers on an
        # unload.  Bleh.
        # https://stackoverflow.com/questions/24681214/unloading-from-redshift-to-s3-ith-headers
        cant_handle_hint(fail_if_cant_handle_hint, 'header-row', hints)
    else:
        quiet_remove(unhandled_hints, 'header-row')

    return redshift_options



from googleapiclient.errors import HttpError
from sqlalchemy.exc import DatabaseError
import tenacity
import logging

logger = logging.getLogger(__name__)


def google_sheets_retry():
    # Example raised when we exceed Google Sheets API rate limits:
    #
    # googleapiclient.errors.HttpError:
    # <HttpError 429 when requesting
    # https://sheets.googleapis.com/v4/spreadsheets/...:clear?alt=json returned "Quota
    # exceeded for quota group 'WriteGroup' and limit 'Write requests per user per 100
    # seconds' of service 'sheets.googleapis.com' for consumer 'project_number:...'.">
    #
    # https://tenacity.readthedocs.io/en/latest/
    #
    # 'max' below is in seconds
    return tenacity.retry(wait=tenacity.wait_random_exponential(multiplier=2, max=120),
                          stop=tenacity.stop_after_attempt(10),
                          before_sleep=tenacity.before_sleep_log(logger, logging.WARNING),
                          retry=(tenacity.retry_if_exception_type(HttpError) &
                                 # Wish the http code were exported as a type of the exception:
                                 #
                                 # https://github.com/googleapis/google-api-python-client/blob/master/googleapiclient/errors.py
                                 tenacity.retry_if_exception(lambda e: 'Quota exceeded' in str(e))),
                          reraise=True)


def bigquery_retry():

    def is_bigquery_rate_limit_exception(e: Exception) -> bool:
        return ('403 Exceeded rate limits:' in str(e) or
                'Job exceeded rate limits' in str(e))

    # Example raised when we exceed BigQuery API rate limits:

    # sqlalchemy.exc.DatabaseError: (google.cloud.bigquery.dbapi.exceptions.DatabaseError)
    # 403 Exceeded rate limits: too many table update operations for this table. For more
    # information, see https://cloud.google.com/bigquery/troubleshooting-errors
    return tenacity.retry(wait=tenacity.wait_random_exponential(multiplier=1, max=60),
                          stop=tenacity.stop_after_attempt(5),
                          before_sleep=tenacity.before_sleep_log(logger, logging.WARNING),
                          retry=(tenacity.retry_if_exception_type(DatabaseError) &
                                 # Wish the specific error were tied to the type or a field...
                                 #
                                 # https://google-cloud-python.readthedocs.io/en/0.32.0/_modules/google/cloud/bigquery/dbapi/exceptions.html#DatabaseError
                                 tenacity.retry_if_exception(is_bigquery_rate_limit_exception)),
                          reraise=True)



from ...utils import quiet_remove
from ...records.delimited import cant_handle_hint
from typing import Set
from ...records.load_plan import RecordsLoadPlan
from ...records.records_format import (
    DelimitedRecordsFormat, ParquetRecordsFormat, AvroRecordsFormat
)
from records_mover.records.delimited import ValidatedRecordsHints
from records_mover.mover_types import _assert_never
from google.cloud.bigquery.job import CreateDisposition, WriteDisposition
from google.cloud import bigquery
import logging

logger = logging.getLogger(__name__)


def load_job_config(unhandled_hints: Set[str],
                    load_plan: RecordsLoadPlan) -> bigquery.LoadJobConfig:

    # https://cloud.google.com/bigquery/docs/loading-data-cloud-storage-parquet#type_conversions
    # https://googleapis.github.io/google-cloud-python/latest/bigquery/generated/google.cloud.bigquery.job.LoadJobConfig.html
    # https://cloud.google.com/bigquery/docs/reference/rest/v2/JobConfiguration#JobConfigurationTableCopy

    fail_if_cant_handle_hint = load_plan.processing_instructions.fail_if_cant_handle_hint
    config = bigquery.LoadJobConfig()

    # clustering_fields: Fields defining clustering for the table
    #
    # > Currently, BigQuery supports clustering over a partitioned
    # > table.  Use clustering over a partitioned table when:
    # > * Your data is already partitioned on a date or timestamp column.
    # > * You commonly use filters or aggregation against
    # >   particular columns in your queries.
    #
    # https://cloud.google.com/bigquery/docs/clustered-tables
    config.clustering_fields = None

    # autodetect: Automatically infer the schema from a sample of the data.
    # schema: Schema of the destination table.
    # create_disposition: Specifies behavior for creating tables.
    #
    # Rely on prep.py in records/ to create the table.
    config.autodetect = False
    config.create_disposition = CreateDisposition.CREATE_NEVER

    # destination_encryption_configuration: Custom encryption configuration for
    # the destination table.
    #
    # Custom encryption configuration (e.g., Cloud KMS keys) or
    # None if using default encryption.
    config.destination_encryption_configuration = None

    # destination_table_description: Union[str, None] name given to destination
    # table.
    config.destination_table_description = None

    # destination_table_friendly_name: Union[str, None] name given
    # to destination table.
    config.destination_table_friendly_name = None

    # ignore_unknown_values: Ignore extra values not represented
    # in the table schema
    config.ignore_unknown_values = load_plan.processing_instructions.fail_if_row_invalid

    # max_bad_records: Number of invalid rows to ignore.
    if load_plan.processing_instructions.max_failure_rows is not None:
        config.max_bad_records = load_plan.processing_instructions.max_failure_rows
        config.allow_jagged_rows = True
    elif load_plan.processing_instructions.fail_if_row_invalid:
        config.max_bad_records = 0
        config.allow_jagged_rows = False
    else:
        config.max_bad_records = 999999
        config.allow_jagged_rows = True

    # write_disposition: Action that occurs if the destination
    # table already exists.
    #
    # Since prep.py handles whatever policy, the table will be
    # already empty if we don't want to append anyway:
    config.write_disposition = WriteDisposition.WRITE_APPEND

    # time_partitioning: Specifies time-based partitioning for the
    # destination table.

    # use_avro_logical_types: For loads of Avro data, governs whether Avro
    # logical types are converted to their corresponding BigQuery types

    # labels: Labels for the job.
    #
    # This method always returns a dict. To change a job’s labels,
    # modify the dict, then call Client.update_job. To delete a
    # label, set its value to None before updating.

    # schema_update_options: Specifies updates to the destination
    # table schema to allow as a side effect of the load job.
    #
    # Allows the schema of the destination table to be updated as
    # a side effect of the query job. Schema update options are
    # supported in two cases: when writeDisposition is
    # WRITE_APPEND; when writeDisposition is WRITE_TRUNCATE and
    # the destination table is a partition of a table, specified
    # by partition decorators. For normal tables, WRITE_TRUNCATE
    # will always overwrite the schema. One or more of the
    # following values are specified: ALLOW_FIELD_ADDITION: allow
    # adding a nullable field to the
    # schema. ALLOW_FIELD_RELAXATION: allow relaxing a required
    # field in the original schema to nullable.
    config.schema_update_options = None
    fail_if_cant_handle_hint = load_plan.processing_instructions.fail_if_cant_handle_hint
    if isinstance(load_plan.records_format, DelimitedRecordsFormat):
        hints = load_plan.records_format.validate(fail_if_cant_handle_hint=fail_if_cant_handle_hint)
        add_load_job_csv_config(unhandled_hints,
                                hints,
                                fail_if_cant_handle_hint,
                                config)
        return config

    if isinstance(load_plan.records_format, ParquetRecordsFormat):
        config.source_format = 'PARQUET'
        return config

    if isinstance(load_plan.records_format, AvroRecordsFormat):
        config.source_format = 'AVRO'
        # https://cloud.google.com/bigquery/docs/loading-data-cloud-storage-avro#logical_types
        config.use_avro_logical_types = True
        return config

    raise NotImplementedError("Not currently able to load "
                              f"{load_plan.records_format.format_type}")


def add_load_job_csv_config(unhandled_hints: Set[str],
                            hints: ValidatedRecordsHints,
                            fail_if_cant_handle_hint: bool,
                            config: bigquery.LoadJobConfig) -> None:
    # source_format: File format of the data.
    config.source_format = 'CSV'

    # encoding: The character encoding of the data.
    # The supported values are UTF-8 or ISO-8859-1.
    # "UTF-8 or ISO-8859-1"
    #
    if hints.encoding == 'UTF8':
        config.encoding = 'UTF-8'
    else:
        # Currently records hints don't support ISO-8859-1
        cant_handle_hint(fail_if_cant_handle_hint, 'encoding', hints)
    quiet_remove(unhandled_hints, 'encoding')

    # field_delimiter: The separator for fields in a CSV file.
    assert isinstance(hints.field_delimiter, str)
    config.field_delimiter = hints.field_delimiter
    quiet_remove(unhandled_hints, 'field-delimiter')

    # allow_jagged_rows: Allow missing trailing optional columns (CSV only).

    # null_marker: Represents a null value (CSV only)
    #
    # (documentation is mangled for this one, but I assume the default is
    # '' or something sensible, so not messing with it)

    # quote_character: Character used to quote data sections (CSV
    # only).
    #
    # [Optional] The value that is used to quote data sections in
    # a CSV file. BigQuery converts the string to ISO-8859-1
    # encoding, and then uses the first byte of the encoded string
    # to split the data in its raw, binary state. The default
    # value is a double-quote ('"'). If your data does not contain
    # quoted sections, set the property value to an empty
    # string. If your data contains quoted newline characters, you
    # must also set the allowQuotedNewlines property to
    # true.
    #
    # @default "

    # I tried a few combinations and found that when you leave quote_character as the default
    #
    # * Fields quoted with "" are loaded without the surrounding quotes in the
    #   string
    # * "" becomes " in a quoted field
    # * "" stays "" in a non-quoted field
    # * nonnumeric quoting works fine
    # * full quoting works fine

    if hints.quoting is None:
        config.quote_character = ''
    elif hints.quoting == 'all' or hints.quoting == 'minimal' or hints.quoting == 'nonnumeric':
        # allow_quoted_newlines: Allow quoted data containing newline
        # characters (CSV only).

        config.allow_quoted_newlines = True

        assert isinstance(hints.quotechar, str)
        config.quote_character = hints.quotechar
        if hints.doublequote:
            pass
        else:
            cant_handle_hint(fail_if_cant_handle_hint, 'doublequote', hints)

    else:
        _assert_never(hints.quoting)
    quiet_remove(unhandled_hints, 'quoting')
    quiet_remove(unhandled_hints, 'quotechar')
    quiet_remove(unhandled_hints, 'doublequote')

    # No mention of escaping in BigQuery documentation, and in
    # practice backslashes come through without being interpreted.
    if hints.escape is None:
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'escape', hints)
    quiet_remove(unhandled_hints, 'escape')

    # skip_leading_rows: Number of rows to skip when reading data (CSV only).
    if hints.header_row:
        config.skip_leading_rows = 1
    else:
        config.skip_leading_rows = 0
    quiet_remove(unhandled_hints, 'header-row')

    # "When you load CSV or JSON data, values in DATE columns must
    #  use the dash (-) separator and the date must be in the
    # following format: YYYY-MM-DD (year-month-day)."
    if hints.dateformat == 'YYYY-MM-DD':
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'dateformat', hints)
    quiet_remove(unhandled_hints, 'dateformat')

    # "When you load JSON or CSV data, values in TIMESTAMP columns
    #  must use a dash (-) separator for the date portion of the
    #  timestamp, and the date must be in the following format:
    #  YYYY-MM-DD (year-month-day). The hh:mm:ss
    #  (hour-minute-second) portion of the timestamp must use a
    #  colon (:) separator."
    #
    #
    # To test, log into BigQuery web console and try SQL like this
    #   (assumption is that the same timestamp parser is used during
    #   CSV loads)
    #
    #      select TIMESTAMP("2000-01-02 16:34:56.789012US/Eastern") as a;
    #
    # Tests performed and result displayed on console query:
    #
    # DATE:
    # * 01-02-2019 (rejected):
    # * 01/02/19 (rejected):
    # * 2019-01-01 (accepted): 2019-01-01
    # DATETIME:
    # * 2019-01-01 1:00pm (rejected):
    # * 2019-01-01 1:00:00pm (rejected)
    # * 2019-01-01 1:00PM (rejected):
    # * 2019-01-01 13:00 (rejected):
    # * 2019-01-01 13:00:00 (accepted): 2019-01-01T13:00:00
    # * 2019-01-01 1:00pm US/Eastern (rejected):
    # * 2019-01-01 1:00:00pm US/Eastern (rejected):
    # * 2019-01-01 13:00:00 US/Eastern (rejected):
    # * 2019-01-01 13:00:00 EST (rejected):
    # * 1997-12-17 07:37:16-08 (rejected)
    # * 2019-01-01T13:00:00 (accepted): 2019-01-01T13:00:00
    #
    # TIME:
    # * 1:00pm (rejected):
    # * 1:00:00pm (rejected):
    # * 13:00 (rejected):
    # * 13:00:00 (accepted): 13:00:00
    # * 1:00pm US/Eastern (rejected):
    # * 1:00pm EST (rejected):
    # * 07:37:16-08 (rejected):
    #
    # TIMESTAMP ("Required format is YYYY-MM-DD
    # HH:MM[:SS[.SSSSSS]]", which is BS, as it doesn't specify the
    # timezone format):
    #
    # * 2019-01-01 1:00pm (rejected):
    # * 2019-01-01 1:00:00pm (rejected)
    # * 2019-01-01 1:00PM (rejected):
    # * 2019-01-01 13:00 (rejected):
    # * 2019-01-01 13:00:00 (accepted): 2019-01-01T13:00:00
    # * 2019-01-01 1:00pm US/Eastern (rejected):
    # * 2019-01-01 1:00:00pm US/Eastern (rejected):
    # * 2019-01-01 13:00:00 US/Eastern (rejected):
    # * 2019-01-01 13:00:00 EST (rejected):
    # * 1997-12-17 07:37:16-08 (accepted): 1997-12-17 15:37:16 UTC
    # * 2019-01-01T13:00:00-08 (accepted): 2019-01-01 21:00:00 UTC
    # * 2000-01-02 16:34:56.789012+0000 (rejected)
    # * 2000-01-02 16:34:56.789012+00:00 (accepted)
    # * 2000-01-02 16:34:56.789012EST (rejected)
    # * 2000-01-02 16:34:56.789012US/Eastern (rejected)
    # * 2000-01-02 16:34:56.789012UTC (accepted): 2000-01-02 16:34:56.789012 UTC
    # * 2000-01-02 16:34:56.789012 UTC (accepted: 2000-01-02 16:34:56.789012 UTC
    #
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#examples
    #
    # https://stackoverflow.com/questions/47466296/bigquery-datetime-format-csv-to-bigquery-yyyy-mm-dd-hhmmss-ssssss
    #
    # BigQuery supports exactly one format of ingesting timestamps
    # with timezones (what they call 'TIMESTAMP' they call timestamps
    # without timezones 'DATETIME'.
    #
    # That format they accept is ISO 8601, which sounds all nice and
    # standardy. Usable timestamps look like 2000-01-02
    # 16:34:56.789012+00:00.
    # Cool cool. The only issue is that Python's strftime doesn't
    # actually provide a way to add the ':' in the timezone
    # offset. The only timezone offset code, %z, does not provide the
    # colon. Other implementations (GNU libc) offers the %:z option,
    # but that doesn't exist in Python and thus in Pandas.
    #
    # So if you're using Python to export timestamps with timezones,
    # you should probably use the `YYYY-MM-DD HH24:MI:SS` format and
    # express them in UTC.
    #
    # https://stackoverflow.com/questions/44836581/does-python-time-strftime-process-timezone-options-correctly-for-rfc-3339
    # https://stackoverflow.com/questions/28729212/pandas-save-date-in-iso-format
    #
    if hints.datetimeformat in ['YYYY-MM-DD HH24:MI:SS', 'YYYY-MM-DD HH:MI:SS']:
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', hints)
    quiet_remove(unhandled_hints, 'datetimeformat')

    if hints.datetimeformattz in ['YYYY-MM-DD HH:MI:SSOF',
                                  'YYYY-MM-DD HH24:MI:SSOF',
                                  'YYYY-MM-DD HH:MI:SS']:
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformattz', hints)
    quiet_remove(unhandled_hints, 'datetimeformattz')

    if hints.timeonlyformat in ['HH24:MI:SS', 'HH:MI:SS']:
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'timeonlyformat', hints)
    quiet_remove(unhandled_hints, 'timeonlyformat')

    # No options to change this.  Tested with unix newlines, dos
    # newlines and mac newlines and all were understood.:
    if hints.record_terminator in ['\n', '\r\n', '\r', None]:
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'record-terminator', hints)
    quiet_remove(unhandled_hints, 'record-terminator')

    # No way to flag compression, but tested uncompressed, with
    # gzip and works great.  .bz2 gives "400 Unsupported
    # compression type".  Not sure about .lzo, but pandas can't
    # handle it regardless, so doubt it's handled.
    if hints.compression is None or hints.compression == 'GZIP':
        pass
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)
    quiet_remove(unhandled_hints, 'compression')



import chardet
from contextlib import contextmanager
from . import PartialRecordsHints
from .csv_streamer import stream_csv, python_encoding_from_hint
import io
import csv
import gzip
import bz2
from .types import HintEncoding, HintRecordTerminator, HintQuoting, HintCompression
from .conversions import hint_encoding_from_chardet
from typing import List, IO, Optional, Iterator, Dict
from records_mover.utils.rewound_fileobj import rewound_fileobj
from records_mover.mover_types import _assert_never
import logging


logger = logging.getLogger(__name__)

HINT_INFERENCE_SAMPLING_SIZE_BYTES = 1024


@contextmanager
def rewound_decompressed_fileobj(fileobj: IO[bytes],
                                 compression: HintCompression) -> Iterator[IO[bytes]]:
    with rewound_fileobj(fileobj) as fileobj_after_rewind:
        if compression is None:
            yield fileobj
        elif compression == 'GZIP':
            yield gzip.GzipFile(mode='rb', fileobj=fileobj_after_rewind)  # type: ignore
        elif compression == 'LZO':
            # This might be useful to implement this:
            #  https://github.com/ir193/python-lzo/blob/master/lzo.py#L44
            raise NotImplementedError('Records mover does not currently know how '
                                      'to decompress LZO files for inspection')
        elif compression == 'BZIP':
            yield bz2.BZ2File(mode='rb', filename=fileobj_after_rewind)
        else:
            _assert_never(compression)


def infer_newline_format(fileobj: IO[bytes],
                         encoding_hint: HintEncoding,
                         compression: HintCompression) ->\
        Optional[HintRecordTerminator]:
    with rewound_decompressed_fileobj(fileobj, compression) as fileobj:
        python_encoding = python_encoding_from_hint[encoding_hint]
        text_fileobj = io.TextIOWrapper(fileobj, encoding=python_encoding)
        try:
            if text_fileobj.newlines is None:  # ...and it almost certainly will be...
                text_fileobj.readline()  # read enough to know newline format
            # https://www.python.org/dev/peps/pep-0278/
            if text_fileobj.newlines is not None:
                logger.info(f"Inferred record terminator as {repr(text_fileobj.newlines)}")
                return str(text_fileobj.newlines)
            else:
                logger.warning("Python could not determine newline format of file.")
                return None
        finally:
            text_fileobj.detach()


def sniff_encoding_hint(fileobj: IO[bytes]) -> Optional[HintEncoding]:
    with rewound_fileobj(fileobj) as fileobj:
        detector = chardet.UniversalDetector()
        while True:
            chunk = fileobj.read(HINT_INFERENCE_SAMPLING_SIZE_BYTES)
            detector.feed(chunk)
            if detector.done or len(chunk) < HINT_INFERENCE_SAMPLING_SIZE_BYTES:
                break
        detector.close()
        assert detector.result is not None
        if 'encoding' in detector.result:
            chardet_encoding = detector.result['encoding']
            if chardet_encoding in hint_encoding_from_chardet:
                return hint_encoding_from_chardet[chardet_encoding]
            else:
                logger.warning("Got unrecognized encoding from chardet "
                               f"sniffing: {detector.result}")
                return None
        else:
            logger.warning(f"Unable to sniff file encoding using chardet: {detector.result}")
            return None


def csv_hints_from_python(fileobj: IO[bytes],
                          record_terminator_hint: Optional[HintRecordTerminator],
                          encoding_hint: HintEncoding,
                          compression: HintCompression) -> PartialRecordsHints:
    # https://docs.python.org/3/library/csv.html#csv.Sniffer
    with rewound_decompressed_fileobj(fileobj,
                                      compression) as fileobj:
        # Sniffer tries to determine quotechar, doublequote,
        # delimiter, skipinitialspace.  does not try to determine
        # lineterminator.
        # https://github.com/python/cpython/blob/master/Lib/csv.py#L165
        python_encoding = python_encoding_from_hint[encoding_hint]
        #
        # TextIOWrapper can only handle standard newline types:
        #
        # https://docs.python.org/3/library/io.html#io.TextIOWrapper
        #
        if record_terminator_hint not in [None, '\n', '\r', '\r\n']:
            logger.info("Unable to infer file with non-standard newlines "
                        f"using Python csv.Sniffer {repr(record_terminator_hint)}.")
            return {}
        text_fileobj = io.TextIOWrapper(fileobj,
                                        encoding=python_encoding,
                                        newline=record_terminator_hint)
        try:
            sniffer = csv.Sniffer()
            sample = text_fileobj.read(HINT_INFERENCE_SAMPLING_SIZE_BYTES)
            #
            # the CSV sniffer's has_header() method seems to only
            # cope with DOS and UNIX newlines, not Mac.  So let's give it
            # UNIX newlines if we know enough to translate, since
            # we're not using it to sniff newline format anyway.
            #
            if record_terminator_hint is not None and record_terminator_hint != '\n':
                sample_with_unix_newlines = sample.replace(record_terminator_hint, '\n')
            else:
                sample_with_unix_newlines = sample
            dialect = sniffer.sniff(sample_with_unix_newlines)
            header_row = sniffer.has_header(sample_with_unix_newlines)
            out: PartialRecordsHints = {
                'doublequote': True if dialect.doublequote else False,
                'field-delimiter': dialect.delimiter,
                'header-row': True if header_row else False,
            }
            if dialect.quotechar is not None:
                out['quotechar'] = dialect.quotechar
            logger.info(f"Python csv.Dialect sniffed: {out}")
            return out
        except csv.Error as e:
            if str(e) == 'Could not determine delimiter':
                logger.info(f"Error from csv.Sniffer--potential single-field file: {str(e)}")
                return {}
            else:
                logger.info(f"Error from csv.Sniffer--potential single-field file: {str(e)}")
                raise
        finally:
            text_fileobj.detach()


def csv_hints_from_pandas(fileobj: IO[bytes],
                          streaming_hints: PartialRecordsHints) -> PartialRecordsHints:
    import pandas

    def attempt_parse(quoting: HintQuoting) -> PartialRecordsHints:
        with rewound_fileobj(fileobj) as fresh_fileobj:
            current_hints = streaming_hints.copy()
            current_hints['quoting'] = quoting
            logger.info(f"Attempting to parse with quoting: {quoting}")
            with stream_csv(fresh_fileobj, current_hints):
                return {
                    'quoting': quoting
                }

    if 'quoting' in streaming_hints:
        return attempt_parse(streaming_hints['quoting'])
    else:
        # Pandas seems to parse quoting=minimal files just fine when
        # you pass in quoting=all, making this technique useless to
        # distinguish between minimal/nonnumeric/all, so we'll only
        # try None and minimal here.
        try:
            return attempt_parse(quoting='minimal')
        except (pandas.errors.ParserError, pandas.errors.EmptyDataError):
            return attempt_parse(quoting=None)


def sniff_compression_hint(fileobj: IO[bytes]) -> HintCompression:
    print(f'Sniffing compression')
    with rewound_fileobj(fileobj) as fileobj_rewound:
        # https://stackoverflow.com/a/13044946/9795956
        magic_dict: Dict[bytes, HintCompression] = {
            b"\x1f\x8b\x08": "GZIP",
            b"\x42\x5a\x68": "BZIP",
            # "\x50\x4b\x03\x04": "zip"
        }

        max_len = max(len(x) for x in magic_dict)

        file_start = fileobj_rewound.read(max_len)
        for magic, filetype in magic_dict.items():
            if file_start.startswith(magic):
                return filetype
        return None


def sniff_hints_from_fileobjs(fileobjs: List[IO[bytes]],
                              initial_hints: PartialRecordsHints) -> PartialRecordsHints:
    if len(fileobjs) != 1:
        # https://github.com/bluelabsio/records-mover/issues/84
        raise NotImplementedError('Cannot currently sniff hints from multiple '
                                  'files--please provide hints')
    fileobj = fileobjs[0]
    hints = sniff_hints(fileobj, initial_hints=initial_hints)
    return hints


def sniff_hints(fileobj: IO[bytes],
                initial_hints: PartialRecordsHints) -> PartialRecordsHints:
    # Major limitations:
    #
    #  * If fileobj isn't rewindable, we can't sniff or we'd keep you
    #    from being able to use later.
    #  * We can't sniff from LZO files yet.
    #  * No detection of 'escape' or date/time format hints.
    #  * Only limited detection of 'quoting' hint.
    try:
        #
        # We'll need to determine compression and encoding to be able
        # to convert from bytes to characters and figure out the rest:
        #
        if 'compression' not in initial_hints:
            compression_hint = sniff_compression_hint(fileobj)
        else:
            compression_hint = initial_hints['compression']
        if 'encoding' not in initial_hints:
            encoding_hint = sniff_encoding_hint(fileobj)
        else:
            encoding_hint = initial_hints['encoding']
        # If guessing was inconclusive, default to UTF8
        final_encoding_hint: HintEncoding = (encoding_hint or 'UTF8')

        #
        # Now we can figure out what type of newlines are used in the
        # file...
        #
        record_terminator_hint: Optional[HintRecordTerminator] = None
        if 'record-terminator' in initial_hints:
            record_terminator_hint = initial_hints['record-terminator']
        else:
            record_terminator_hint = infer_newline_format(fileobj,
                                                          final_encoding_hint,
                                                          compression_hint)

        #
        # Now we have enough to study each line of the file, Python's
        # csv.Sniffer() can teach us some things about how each field
        # is encoded and whether there's a header:
        #
        other_inferred_csv_hints = {}
        if record_terminator_hint is not None:
            other_inferred_csv_hints['record-terminator'] = record_terminator_hint
            python_inferred_hints = csv_hints_from_python(fileobj,
                                                          record_terminator_hint,
                                                          final_encoding_hint,
                                                          compression_hint)
        else:
            python_inferred_hints = {}

        #
        # Pandas can both validate that we chose correctly by parsing
        # the file using what we have so far, and give us a crude shot
        # at finding a working 'quoting' hint:
        #
        streaming_hints = initial_hints.copy()
        streaming_hints['compression'] = compression_hint
        if encoding_hint is not None:
            streaming_hints['encoding'] = encoding_hint
        streaming_hints.update(python_inferred_hints)
        pandas_inferred_hints = csv_hints_from_pandas(fileobj, streaming_hints)

        #
        # Let's combine these together and present back a refined
        # version of the initial hints:
        #
        out: PartialRecordsHints = {
            'compression': compression_hint,
            **pandas_inferred_hints,  # type: ignore
            **python_inferred_hints,
            'encoding': final_encoding_hint,
            **other_inferred_csv_hints,
            **initial_hints
        }
        logger.info(f"Inferred hints from combined sources: {out}")
        return out
    except OSError:
        logger.warning("Could not sniff hints, as stream is not rewindable")
        return {}
    except NotImplementedError as e:
        logger.warning(f"Could not sniff hints due to current limitations in records mover: {e}")
        return {}



from contextlib import contextmanager
from sqlalchemy_redshift.commands import CopyCommand
from records_mover.logging import register_secret
from ..loader import LoaderFromRecordsDirectory
from ...records.records_directory import RecordsDirectory
from ...records.records_format import BaseRecordsFormat, DelimitedRecordsFormat, AvroRecordsFormat
from ...records.processing_instructions import ProcessingInstructions
import sqlalchemy
from sqlalchemy.schema import Table
import logging
from .records_copy import redshift_copy_options
from ...records.load_plan import RecordsLoadPlan
from ..errors import CredsDoNotSupportS3Import, NoTemporaryBucketConfiguration
from typing import Optional, Union, List, Iterator
from ...url import BaseDirectoryUrl
from botocore.credentials import Credentials
from ...records.delimited import complain_on_unhandled_hints

logger = logging.getLogger(__name__)


class RedshiftLoader(LoaderFromRecordsDirectory):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                 meta: sqlalchemy.MetaData,
                 s3_temp_base_loc: Optional[BaseDirectoryUrl])\
            -> None:
        self.db = db
        self.meta = meta
        self.s3_temp_base_loc = s3_temp_base_loc

    @contextmanager
    def temporary_s3_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        if self.s3_temp_base_loc is None:
            raise NoTemporaryBucketConfiguration('Please provide a scratch S3 URL in your config '
                                                 '(e.g., set SCRATCH_S3_URL to an s3:// URL)')
        else:
            with self.s3_temp_base_loc.temporary_directory() as temp_loc:
                yield temp_loc

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> Optional[int]:

        if directory.scheme != 's3':
            with self.temporary_s3_directory_loc() as temp_s3_loc:
                s3_directory = directory.copy_to(temp_s3_loc)
                return self.load(schema=schema,
                                 table=table,
                                 load_plan=load_plan,
                                 directory=s3_directory)

        to = Table(table, self.meta, schema=schema)  # no autoload
        unhandled_hints = set()
        if isinstance(load_plan.records_format, DelimitedRecordsFormat):
            unhandled_hints = set(load_plan.records_format.hints.keys())
        processing_instructions = load_plan.processing_instructions
        redshift_options = redshift_copy_options(unhandled_hints,
                                                 load_plan.records_format,
                                                 processing_instructions.fail_if_cant_handle_hint,
                                                 processing_instructions.fail_if_row_invalid,
                                                 processing_instructions.max_failure_rows)
        logger.info(f"Copying to Redshift with options: {redshift_options}")
        if isinstance(load_plan.records_format, DelimitedRecordsFormat):
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, load_plan.records_format.hints)
        # http://sqlalchemy-redshift.readthedocs.io/en/latest/commands.html
        loc = directory.loc
        if not callable(getattr(loc, 'aws_creds', None)):
            raise NotImplementedError('Redshift can only load from an S3 bucket')
        else:
            aws_creds: Optional[Credentials] = directory.loc.aws_creds()  # type: ignore
            if aws_creds is None:
                raise CredsDoNotSupportS3Import('Please provide AWS credentials '
                                                '(run "aws configure")')
            #
            # Upon error, an exception is raised with the full SQL -
            # including the AWS creds inside.  Let's register those
            # with the logger so they get redacted.
            #
            register_secret(aws_creds.token)
            register_secret(aws_creds.secret_key)

            copy = CopyCommand(to=to, data_location=directory.loc.url + '_manifest',
                               access_key_id=aws_creds.access_key,
                               secret_access_key=aws_creds.secret_key,
                               session_token=aws_creds.token, manifest=True,
                               region=directory.loc.region,  # type: ignore
                               empty_as_null=True,
                               **redshift_options)  # type: ignore
            logger.info(f"Starting Redshift COPY from {directory}...")
            redshift_pid: int = self.db.execute("SELECT pg_backend_pid();").scalar()
            try:
                self.db.execute(copy)
            except sqlalchemy.exc.InternalError:
                # Upon a load erorr, we receive:
                #
                #  sqlalchemy.exc.InternalError:
                #  (psycopg2.errors.InternalError_) Load into table 'tablename'
                #    failed.  Check 'stl_load_errors' system table for details.
                logger.warning("Caught load error - "
                               "for details, run this query: "
                               f"SELECT * FROM stl_load_errors WHERE session={redshift_pid}")
                raise
            logger.info("Redshift COPY complete.")
            return None  # redshift doesn't give reliable info on load results

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        try:
            processing_instructions = ProcessingInstructions()
            load_plan = RecordsLoadPlan(records_format=source_records_format,
                                        processing_instructions=processing_instructions)
            unhandled_hints = set()
            records_format = load_plan.records_format
            if isinstance(records_format, DelimitedRecordsFormat):
                unhandled_hints = set(records_format.hints.keys())
            processing_instructions = load_plan.processing_instructions
            redshift_copy_options(unhandled_hints,
                                  records_format,
                                  processing_instructions.fail_if_cant_handle_hint,
                                  processing_instructions.fail_if_row_invalid,
                                  processing_instructions.max_failure_rows)
            if isinstance(records_format, DelimitedRecordsFormat):
                complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                            unhandled_hints,
                                            records_format.hints)
            return True
        except NotImplementedError:
            return False

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return [
            # Redshift is pretty flexible with date/time/timezone
            # parsing, so let's use them and go back to saner defaults
            # than those of MS Excel - e.g., the default csv and bigquery
            # formats don't include timezones.
            DelimitedRecordsFormat(variant='csv',
                                   hints={
                                       'dateformat': 'YYYY-MM-DD',
                                       'timeonlyformat': 'HH24:MI:SS',
                                       'datetimeformat': 'YYYY-MM-DD HH:MI:SS',
                                       'datetimeformattz': 'YYYY-MM-DD HH:MI:SSOF',
                                   }),
            # 'bigquery' and 'csv' support both newlines in strings as
            # well as empty strings (but not timezones)
            DelimitedRecordsFormat(variant='bigquery'),
            DelimitedRecordsFormat(variant='csv'),
            # The default 'bluelabs' format can't represent empty
            # strings when exported via Pandas (see
            # https://github.com/pandas-dev/pandas/issues/15891) - but
            # it can with this flag.  Unfortunately, Redshift doesn't
            # support newlines in strings when using this format.
            DelimitedRecordsFormat(variant='bluelabs', hints={
                'quoting': 'all',
            }),
            # Supports newlines in strings, but not empty strings.
            DelimitedRecordsFormat(variant='bluelabs'),
            AvroRecordsFormat(),
        ]

    def best_scheme_to_load_from(self) -> str:
        return 's3'

    def temporary_loadable_directory_scheme(self) -> str:
        return 's3'

    @contextmanager
    def temporary_loadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        with self.temporary_s3_directory_loc() as temp_loc:
            yield temp_loc

    def has_temporary_loadable_directory_loc(self) -> bool:
        return self.s3_temp_base_loc is not None



from .constraints import RecordsSchemaFieldConstraints  # noqa
from .string import RecordsSchemaFieldStringConstraints  # noqa
from .integer import RecordsSchemaFieldIntegerConstraints  # noqa
from .decimal import RecordsSchemaFieldDecimalConstraints  # noqa
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .constraints import FieldConstraintsDict  # noqa



from contextlib import contextmanager
from sqlalchemy_redshift.commands import UnloadFromSelect
from ...records.records_directory import RecordsDirectory
import sqlalchemy
from sqlalchemy.sql import text
from sqlalchemy.schema import Table
import logging
from records_mover.db.quoting import quote_schema_and_table
from records_mover.logging import register_secret
from .records_unload import redshift_unload_options
from ...records.unload_plan import RecordsUnloadPlan
from ...records.records_format import (
    BaseRecordsFormat, DelimitedRecordsFormat, ParquetRecordsFormat
)
from typing import Union, Callable, Optional, List, Iterator
from ...url.base import BaseDirectoryUrl
from botocore.credentials import Credentials
from ..errors import CredsDoNotSupportS3Export, NoTemporaryBucketConfiguration
from ...records.delimited import complain_on_unhandled_hints
from ..unloader import Unloader


logger = logging.getLogger(__name__)


class RedshiftUnloader(Unloader):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                 table: Callable[[str, str], Table],
                 s3_temp_base_loc: Optional[BaseDirectoryUrl],
                 **kwargs) -> None:
        super().__init__(db=db)
        self.table = table
        self.s3_temp_base_loc = s3_temp_base_loc

    @contextmanager
    def temporary_unloadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        with self.temporary_s3_directory_loc() as temp_loc:
            yield temp_loc

    @contextmanager
    def temporary_s3_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        if self.s3_temp_base_loc is None:
            raise NoTemporaryBucketConfiguration('Please provide a scratch S3 URL in your config '
                                                 '(e.g., set SCRATCH_S3_URL to an s3:// URL)')
        else:
            with self.s3_temp_base_loc.temporary_directory() as temp_loc:
                yield temp_loc

    def unload_to_s3_directory(self,
                               schema: str,
                               table: str,
                               unload_plan: RecordsUnloadPlan,
                               directory: RecordsDirectory) -> Optional[int]:
        logger.info(f"Starting Redshift unload to {directory.loc} as "
                    f"{unload_plan.records_format}...")
        unhandled_hints = set()
        if isinstance(unload_plan.records_format, DelimitedRecordsFormat):
            unhandled_hints = set(unload_plan.records_format.hints.keys())
        processing_instructions = unload_plan.processing_instructions
        redshift_options = redshift_unload_options(unhandled_hints,
                                                   unload_plan.records_format,
                                                   processing_instructions.fail_if_cant_handle_hint)
        if isinstance(unload_plan.records_format, DelimitedRecordsFormat):
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, unload_plan.records_format.hints)
        # http://sqlalchemy-redshift.readthedocs.io/en/latest/commands.html
        loc = directory.loc
        if not callable(getattr(loc, 'aws_creds')):
            raise NotImplementedError('Redshift can only load from an S3 bucket')
        else:
            aws_creds: Optional[Credentials] = loc.aws_creds()  # type: ignore
            if aws_creds is None:
                raise CredsDoNotSupportS3Export('Please provide AWS credentials '
                                                '(run "aws configure")')
            #
            # Upon error, an exception is raised with the full SQL -
            # including the AWS creds inside.  Let's register those
            # with the logger so they get redacted.
            #
            register_secret(aws_creds.token)
            register_secret(aws_creds.secret_key)
            select = text(f"SELECT * FROM {quote_schema_and_table(self.db, schema, table)}")
            unload = UnloadFromSelect(select=select,
                                      access_key_id=aws_creds.access_key,
                                      secret_access_key=aws_creds.secret_key,
                                      session_token=aws_creds.token, manifest=True,
                                      unload_location=directory.loc.url, **redshift_options)
            try:
                self.db.execute(unload)
                out = self.db.execute(text("SELECT pg_last_unload_count()"))
                rows: Optional[int] = out.scalar()
                assert rows is not None
                logger.info(f"Just unloaded {rows} rows")
                out.close()
            except sqlalchemy.exc.DatabaseError as e:
                if 'SSL SYSCALL error: Operation timed out' in str(e):
                    # Large database UNLOADs can take hours, and it's
                    # likely the connection between the client and
                    # server will get disconnected.  In this
                    # circumstance, we want to wait until unload is
                    # complete and then proceed.
                    logger.info("Server disconnected - awaiting completion "
                                "of work by checking S3 bucket...")
                    directory.await_completion(log_level=logging.DEBUG,
                                               ms_between_polls=5000,
                                               manifest_filename='manifest')
                    # pg_last_unload_count() doesn't work
                    # after this situation - the query shows up as
                    rows = None
                    logger.info("Unload complete.")
                else:
                    raise

            return rows

    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> Optional[int]:
        if directory.scheme == 's3':
            s3_directory = directory
            return self.unload_to_s3_directory(schema, table, unload_plan, s3_directory)
        else:
            with self.temporary_s3_directory_loc() as temp_s3_loc:
                s3_directory = RecordsDirectory(records_loc=temp_s3_loc)
                out = self.unload_to_s3_directory(schema, table, unload_plan, s3_directory)
                directory.copy_from(temp_s3_loc)
                return out

    def can_unload_to_scheme(self, scheme: str) -> bool:
        if scheme == 's3':
            return True
        # Otherwise we'll need a temporary bucket configured for
        # Redshift to unload into
        return self.s3_temp_base_loc is not None

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        return [DelimitedRecordsFormat(variant='bluelabs'), ParquetRecordsFormat()]

    def can_unload_format(self, target_records_format: BaseRecordsFormat) -> bool:
        try:
            unload_plan = RecordsUnloadPlan(records_format=target_records_format)
            unhandled_hints = set()
            if isinstance(unload_plan.records_format, DelimitedRecordsFormat):
                unhandled_hints = set(unload_plan.records_format.hints.keys())
            processing_instructions = unload_plan.processing_instructions
            redshift_unload_options(unhandled_hints,
                                    unload_plan.records_format,
                                    processing_instructions.fail_if_cant_handle_hint)
            if isinstance(unload_plan.records_format, DelimitedRecordsFormat):
                complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                            unhandled_hints,
                                            unload_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False



import sqlalchemy
from sqlalchemy import MetaData
from sqlalchemy.schema import Table
from ..quoting import quote_value
from ...url.resolver import UrlResolver
from ...records.load_plan import RecordsLoadPlan
from ...records.delimited import complain_on_unhandled_hints
from ...records.records_format import DelimitedRecordsFormat, BaseRecordsFormat
from ...records.processing_instructions import ProcessingInstructions
from .sqlalchemy_postgres_copy import copy_from
from .copy_options import postgres_copy_from_options
from typing import IO, Union, List, Iterable
from ..loader import LoaderFromFileobj
import logging

logger = logging.getLogger(__name__)


class PostgresLoader(LoaderFromFileobj):
    def __init__(self,
                 url_resolver: UrlResolver,
                 meta: MetaData,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine]) -> None:
        self.url_resolver = url_resolver
        self.db = db
        self.meta = meta

    def load_from_fileobj(self,
                          schema: str,
                          table: str,
                          load_plan: RecordsLoadPlan,
                          fileobj: IO[bytes]) -> None:
        return self.load_from_fileobjs(schema=schema,
                                       table=table,
                                       load_plan=load_plan,
                                       fileobjs=[fileobj])

    def load_from_fileobjs(self,
                           schema: str,
                           table: str,
                           load_plan: RecordsLoadPlan,
                           fileobjs: Iterable[IO[bytes]]) -> None:
        records_format = load_plan.records_format
        if not isinstance(records_format, DelimitedRecordsFormat):
            raise NotImplementedError("Not currently able to load "
                                      f"{records_format.format_type}")
        processing_instructions = load_plan.processing_instructions
        unhandled_hints = set(records_format.hints.keys())
        date_order_style, postgres_options = postgres_copy_from_options(unhandled_hints, load_plan)
        if date_order_style is None:
            # U-S-A!  U-S-A!
            date_order_style = 'MDY'
        logger.info(f"PostgreSQL load options: {postgres_options}")
        complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                    unhandled_hints,
                                    records_format.hints)

        table_obj = Table(table,
                          self.meta,
                          schema=schema,
                          autoload=True,
                          autoload_with=self.db)

        with self.db.engine.begin() as conn:
            # https://www.postgresql.org/docs/8.3/sql-set.html
            #
            # The effects of SET LOCAL last only till the end of the
            # current transaction, whether committed or not. A special
            # case is SET followed by SET LOCAL within a single
            # transaction: the SET LOCAL value will be seen until the end
            # of the transaction, but afterwards (if the transaction is
            # committed) the SET value will take effect.
            date_style = f"ISO, {date_order_style}"
            sql = f"SET LOCAL DateStyle = {quote_value(conn, date_style)}"
            logger.info(sql)
            conn.execute(sql)

            for fileobj in fileobjs:
                # Postgres COPY FROM defaults to appending data--we
                # let the records Prep class decide what to do about
                # the existing table, so it's safe to call this
                # multiple times and append until done:
                copy_from(fileobj,
                          table_obj,
                          conn,
                          **postgres_options)
        logger.info('Copy complete')

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        try:
            processing_instructions = ProcessingInstructions()
            load_plan = RecordsLoadPlan(records_format=source_records_format,
                                        processing_instructions=processing_instructions)
            if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
                return False

            unhandled_hints = set(load_plan.records_format.hints.keys())
            processing_instructions = load_plan.processing_instructions
            postgres_copy_from_options(unhandled_hints, load_plan)
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, load_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return [
            # To validate that these load without pandas, watch
            # logging while running and verify 'dataframe' doesn't
            # appear:
            #
            # ./itest shell
            #
            # mvrec file2table --source.variant bluelabs
            # --source.no_compression
            # tests/integration/resources/delimited-bluelabs-no-header.csv
            # dockerized-postgres public bluelabsformat
            DelimitedRecordsFormat(variant='bluelabs',
                                   hints={'compression': None}),
            # mvrec file2table --source.variant csv
            # --source.no_compression
            # tests/integration/resources/delimited-csv-with-header.csv
            # dockerized-postgres public csvformat
            DelimitedRecordsFormat(variant='csv',
                                   hints={'compression': None}),
            # mvrec file2table --source.variant bigquery
            # --source.no_compression
            # tests/integration/resources/delimited-bigquery-with-header.csv
            # dockerized-postgres public bigqueryformat
            DelimitedRecordsFormat(variant='bigquery',
                                   hints={'compression': None}),
        ]



from urllib.parse import urlparse
from .base import BaseFileUrl, BaseDirectoryUrl
import inspect
from typing import Callable, Optional, Dict, Type, Union, TYPE_CHECKING
from typing_extensions import TypedDict
import logging
if TYPE_CHECKING:
    import google.cloud.storage  # noqa
    import boto3.session  # noqa


logger = logging.getLogger(__name__)

CreatesUrls = Callable[..., Union[BaseFileUrl, BaseDirectoryUrl]]


directory_url_ctors: Dict[str, Union[Type[BaseDirectoryUrl], CreatesUrls]] = {}
file_url_ctors: Dict[str, Union[Type[BaseFileUrl], CreatesUrls]] = {}


def init_urls() -> None:
    try:
        from .s3.s3_url import S3Url
    except ModuleNotFoundError:
        logger.debug('No S3 support', exc_info=True)
        S3Url = None  # type: ignore
    try:
        from .gcs.gcs_file_url import GCSFileUrl
        from .gcs.gcs_directory_url import GCSDirectoryUrl
    except ModuleNotFoundError:
        logger.debug('No Google Cloud Storage support', exc_info=True)
        GCSFileUrl = None  # type: ignore
        GCSDirectoryUrl = None  # type: ignore
    from .filesystem import FilesystemDirectoryUrl, FilesystemFileUrl
    from .http import HttpFileUrl
    if len(directory_url_ctors) == 0:
        if S3Url is not None:
            directory_url_ctors['s3'] = S3Url
        if GCSDirectoryUrl is not None:
            directory_url_ctors['gs'] = GCSDirectoryUrl
        directory_url_ctors['file'] = FilesystemDirectoryUrl
    if len(file_url_ctors) == 0:
        if S3Url is not None:
            file_url_ctors['s3'] = S3Url
        if GCSFileUrl is not None:
            file_url_ctors['gs'] = GCSFileUrl
        file_url_ctors['file'] = FilesystemFileUrl

        file_url_ctors['http'] = HttpFileUrl
        file_url_ctors['https'] = HttpFileUrl


class UrlClassKwArgs(TypedDict, total=False):
    gcs_client: 'google.cloud.storage.Client'
    gcp_credentials: 'google.auth.credentials.Credentials'
    boto3_session: 'boto3.session.Session'


class UrlResolver:
    def __init__(self,
                 boto3_session_getter:
                 Callable[[],
                          Optional['boto3.session.Session']],

                 gcs_client_getter:
                 Callable[[],
                          Optional['google.cloud.storage.Client']],

                 gcp_credentials_getter:
                 Callable[[],
                          Optional['google.auth.credentials.Credentials']])\
            -> None:
        self.boto3_session_getter = boto3_session_getter
        self.gcs_client_getter = gcs_client_getter
        self.gcp_credentials_getter = gcp_credentials_getter

    def file_url(self, url: str) -> BaseFileUrl:
        init_urls()
        parsed_url = urlparse(url)
        if parsed_url.scheme in file_url_ctors:
            ctor = file_url_ctors[parsed_url.scheme]

            out = ctor(url, **self._kwargs_for_function(ctor))
            if not isinstance(out, BaseFileUrl):
                raise TypeError(f"Not a file url: {url}")
            return out
        else:
            raise NotImplementedError(f"Teach me how to create FileUrls for {parsed_url.scheme}")

    def _kwargs_for_function(self, fn: Callable) -> UrlClassKwArgs:
        parameters: Dict[str, Type] = inspect.signature(fn).parameters
        out: UrlClassKwArgs = {}
        if 'boto3_session' in parameters:
            session = self.boto3_session_getter()
            if session is None:
                raise EnvironmentError('URL requires an AWS boto3 session, but none '
                                       'is configured.  Please configure your credentials.')
            out["boto3_session"] = session
        if 'gcs_client' in parameters:
            gcs_client = self.gcs_client_getter()
            if gcs_client is None:
                raise EnvironmentError('URL requires a GCP client for Google '
                                       'Cloud Storage use, but none are configured.  '
                                       'Please configure your credentials '
                                       '(e.g., "gcloud auth application-default '
                                       'login --project <project>" and set '
                                       'GCP_PROJECT to your project name)')
            out["gcs_client"] = gcs_client
        if 'gcp_credentials' in parameters:
            gcp_credentials = self.gcp_credentials_getter()
            if gcp_credentials is None:
                raise EnvironmentError('URL requires GCP credentials, but none are configured.  '
                                       'Please configure your credentials.')
            out["gcp_credentials"] = gcp_credentials
        return out

    def directory_url(self, url: str) -> BaseDirectoryUrl:
        init_urls()
        parsed_url = urlparse(url)
        if parsed_url.scheme in directory_url_ctors:
            ctor = directory_url_ctors[parsed_url.scheme]
            out = ctor(url, **self._kwargs_for_function(ctor))
            if not isinstance(out, BaseDirectoryUrl):
                raise TypeError(f"Not a directory url: {url}")
            return out
        else:
            raise NotImplementedError("Teach me how to create DirectoryUrls for "
                                      f"{parsed_url.scheme}")



from urllib.parse import urlparse, unquote
import boto3
from botocore.credentials import ReadOnlyCredentials
from typing import TypeVar, Callable, Optional, Any, Union, TYPE_CHECKING
from ..base import BaseDirectoryUrl, BaseFileUrl
if TYPE_CHECKING:
    # http://mypy.readthedocs.io/en/latest/common_issues.html#import-cycles
    from .s3_file_url import S3FileUrl  # noqa
    from .s3_directory_url import S3DirectoryUrl  # noqa

    # These type stubs aren't real classes, so only use their names
    # during type checking
    from boto3.session import S3ResourceTypeStub, S3ClientTypeStub


T = TypeVar('T', bound='S3BaseUrl')


class S3BaseUrl:
    def __init__(self,
                 url: str,
                 boto3_session: boto3.session.Session,
                 S3Url: Callable[[str, Any], Union['S3DirectoryUrl', 'S3FileUrl']]) -> None:
        parsed = urlparse(url)
        self.scheme = parsed.scheme
        self.url = url
        # https://docs.python.org/2/library/urlparse.html
        #
        # "The components are not broken up in smaller parts (for
        # example, the network location is a single string), and %
        # escapes are not expanded."
        self.key = unquote(parsed.path[1:])
        self.bucket = parsed.netloc
        self.region = boto3_session.region_name
        self._boto3_session = boto3_session
        self.s3_resource: S3ResourceTypeStub = boto3_session.resource('s3')
        self.s3_client: S3ClientTypeStub = boto3_session.client('s3')
        self.S3Url = S3Url

    def _directory(self, url: str) -> 'S3DirectoryUrl':
        out = self.S3Url(url, self._boto3_session)
        if not isinstance(out, BaseDirectoryUrl):
            raise TypeError(f"Not a directory url: {url}")
        return out

    def _file(self, url: str) -> 'S3FileUrl':
        out = self.S3Url(url, self._boto3_session)
        if not isinstance(out, BaseFileUrl):
            raise TypeError(f"Not a file url: {url}")
        return out

    def aws_creds(self) -> Optional[ReadOnlyCredentials]:
        """Pull a boto3 credential object that can access this directory.
        Note that these may be expiring and auto-refreshing instance
        creds, so be careful with the result - use the attributes off
        of a single instance rather than calling multiple times, and
        don't save it for later - pull a fresh version when you need
        it.
        """
        creds = self._boto3_session.get_credentials()
        if creds is None:
            return None
        return creds.get_frozen_credentials()

    def containing_directory(self) -> BaseDirectoryUrl:
        parent_url = '/'.join(self.url.split('/')[:-1]) + '/'
        return self._directory(parent_url)

    def _key_in_same_bucket(self, key: str) -> BaseFileUrl:
        return self._file(f"s3://{self.bucket}/{key}")

    def is_directory(self) -> bool:
        return self.url.endswith('/')

    def directory_in_this_directory(self: T, directory_name: str) -> BaseDirectoryUrl:
        raise NotImplementedError()



from records_mover.db.quoting import quote_value
import sqlalchemy
from contextlib import contextmanager
from ...records.unload_plan import RecordsUnloadPlan
from ...records.records_format import BaseRecordsFormat, DelimitedRecordsFormat
from ...url.base import BaseDirectoryUrl
from ..errors import (NoTemporaryBucketConfiguration, LoadUnloadError,
                      CredsDoNotSupportS3Export, DatabaseDoesNotSupportS3Export)
from ...records.records_directory import RecordsDirectory
from .export_sql import vertica_export_sql
from .records_export_options import vertica_export_options
from ...records.delimited import complain_on_unhandled_hints
from ..unloader import Unloader
import logging
from typing import Iterator, Optional, Union, List, TYPE_CHECKING
if TYPE_CHECKING:
    from botocore.credentials import Credentials  # noqa


logger = logging.getLogger(__name__)


class VerticaUnloader(Unloader):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine],
                 s3_temp_base_loc: Optional[BaseDirectoryUrl]) -> None:
        super().__init__(db=db)
        self.s3_temp_base_loc = s3_temp_base_loc

    @contextmanager
    def temporary_unloadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        if self.s3_temp_base_loc is None:
            raise NoTemporaryBucketConfiguration('Please provide a scratch S3 URL in your config '
                                                 '(e.g., set SCRATCH_S3_URL to an s3:// URL)')
        else:
            with self.s3_temp_base_loc.temporary_directory() as temp_loc:
                yield temp_loc

    def aws_creds_sql(self, aws_id: str, aws_secret: str) -> str:
        return """
            ALTER SESSION SET UDPARAMETER FOR awslib aws_id={aws_id};
            ALTER SESSION SET UDPARAMETER FOR awslib aws_secret={aws_secret};
        """.format(aws_id=quote_value(self.db.engine, aws_id),
                   aws_secret=quote_value(self.db.engine, aws_secret))

    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> Optional[int]:
        if not self.s3_export_available():
            raise NotImplementedError('S3 currently required for Vertica bulk unload')
        try:
            if directory.scheme == 's3':
                return self.unload_to_s3_directory(schema, table, unload_plan, directory)
            else:
                with self.temporary_unloadable_directory_loc() as temp_s3_loc:
                    s3_directory = RecordsDirectory(records_loc=temp_s3_loc)
                    out = self.unload_to_s3_directory(schema, table, unload_plan, s3_directory)
                    directory.copy_from(temp_s3_loc)
                    return out
        except LoadUnloadError as e:
            msg = str(e)
            if msg:
                logger.warning(msg)
            raise NotImplementedError('S3 currently required for Vertica bulk unload')

    def s3_temp_bucket_available(self) -> bool:
        return self.s3_temp_base_loc is not None

    def s3_export_available(self) -> bool:
        out = self.db.execute("SELECT lib_name from user_libraries where lib_name = 'awslib'")
        available = len(list(out.fetchall())) == 1
        if not available:
            logger.info("Not attempting S3 export - no access to awslib in Vertica")
        return available

    def unload_to_s3_directory(self,
                               schema: str,
                               table: str,
                               unload_plan: RecordsUnloadPlan,
                               directory: RecordsDirectory) -> int:
        if not isinstance(unload_plan.records_format, DelimitedRecordsFormat):
            raise NotImplementedError("This only supports delimited mode for now")

        loc = directory.loc
        if not callable(getattr(loc, 'aws_creds')):
            raise NotImplementedError('No AWS creds loaded into location')

        aws_creds: Optional['Credentials'] = directory.loc.aws_creds()  # type: ignore
        if aws_creds is None:
            raise CredsDoNotSupportS3Export('Please provide AWS credentials (run "aws configure")')
        if aws_creds.token is not None:
            raise CredsDoNotSupportS3Export("Vertica does not support AWS security tokens--"
                                            "please create a dedicated credential")
        unhandled_hints = set(unload_plan.records_format.hints.keys())
        processing_instructions = unload_plan.processing_instructions
        try:
            s3_sql = self.aws_creds_sql(aws_creds.access_key, aws_creds.secret_key)
            self.db.execute(s3_sql)
        except sqlalchemy.exc.ProgrammingError as e:
            raise DatabaseDoesNotSupportS3Export(str(e))

        vertica_options = vertica_export_options(unhandled_hints, unload_plan)
        complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                    unhandled_hints,
                                    unload_plan.records_format.hints)

        export_sql = vertica_export_sql(db_engine=self.db.engine,
                                        table=table,
                                        schema=schema,
                                        s3_url=directory.loc.url,
                                        **vertica_options)
        logger.info(export_sql)
        export_result = self.db.execute(export_sql).fetchall()
        directory.save_preliminary_manifest()
        export_count = 0
        for record in export_result:
            export_count += int(record.rows)
        return export_count

    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        return [DelimitedRecordsFormat(variant='vertica')]

    def can_unload_to_scheme(self, scheme: str) -> bool:
        if not self.s3_export_available():
            return False
        return scheme == 's3' or self.s3_temp_bucket_available()

    def can_unload_format(self, target_records_format: BaseRecordsFormat) -> bool:
        try:
            unload_plan = RecordsUnloadPlan(records_format=target_records_format)
            if not isinstance(unload_plan.records_format, DelimitedRecordsFormat):
                return False

            unhandled_hints = set(unload_plan.records_format.hints.keys())
            processing_instructions = unload_plan.processing_instructions
            vertica_export_options(unhandled_hints, unload_plan)
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints,
                                        unload_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False



from enum import Enum
import math
from typing import Optional

INT8_MAX = 127
INT8_MIN = -128
UINT8_MAX = 255
UINT8_MIN = 0
INT16_MAX = 32767
INT16_MIN = -32768
UINT16_MAX = 65535
UINT16_MIN = 0
INT24_MAX = 8388607
INT24_MIN = -8388608
UINT24_MAX = 16777215
UINT24_MIN = 0
INT32_MAX = 2147483647
INT32_MIN = -2147483648
UINT32_MAX = 4294967295
UINT32_MIN = 0
INT64_MAX = 9223372036854775807
INT64_MIN = -9223372036854775808
UINT64_MAX = 18446744073709551615
UINT64_MIN = 0
FLOAT16_SIGNIFICAND_BITS = 11
FLOAT32_SIGNIFICAND_BITS = 23
FLOAT64_SIGNIFICAND_BITS = 53
FLOAT80_SIGNIFICAND_BITS = 64


class IntegerType(Enum):
    INT8 = (INT8_MIN, INT8_MAX)
    UINT8 = (UINT8_MIN, UINT8_MAX)
    INT16 = (INT16_MIN, INT16_MAX)
    UINT16 = (UINT16_MIN, UINT16_MAX)
    INT24 = (INT24_MIN, INT24_MAX)
    UINT24 = (UINT24_MIN, UINT24_MAX)
    INT32 = (INT32_MIN, INT32_MAX)
    UINT32 = (UINT32_MIN, UINT32_MAX)
    INT64 = (INT64_MIN, INT64_MAX)
    UINT64 = (UINT64_MIN, UINT64_MAX)

    def __init__(self, min_: int, max_: int):
        self.min_ = min_
        self.max_ = max_

    def is_cover_for(self, low_value: int, high_value: int) -> bool:
        return low_value >= self.min_ and high_value <= self.max_

    @property
    def range(self):
        return (self.min_, self.max_)

    @classmethod
    def smallest_cover_for(cls, low_value: int, high_value: int) -> Optional['IntegerType']:
        for int_type in cls:
            if int_type.is_cover_for(low_value, high_value):
                return int_type
        return None


# https://stackoverflow.com/questions/2189800/length-of-an-integer-in-python
def num_digits(n: int) -> int:
    if n > 0:
        digits = int(math.log10(n)) + 1
    elif n == 0:
        digits = 1
    else:
        digits = int(math.log10(-n)) + 2  # +1 if you don't count the '-'
    return digits



from db_facts.db_facts_types import DBFacts
from .database import db_facts_from_env
from typing import TYPE_CHECKING, Iterable, Union, Optional, Dict, Any
from records_mover.mover_types import PleaseInfer
from config_resolver import get_config
import os
import logging
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa
    import boto3  # noqa


logger = logging.getLogger(__name__)

_GSHEETS_SCOPES = ('https://www.googleapis.com/auth/spreadsheets',)

_GCS_SCOPES = ('https://www.googleapis.com/auth/devstorage.full_control',
               'https://www.googleapis.com/auth/devstorage.read_only',
               'https://www.googleapis.com/auth/devstorage.read_write')


# this interfaces here are probably unstable until we figure out how
# best to integrate airflow hooks and connections (and maybe
# Kubernetes secrets) in with it.  The current (unfinalized and
# untested) thought is to use the *_creds_name argument to reflect the
# Airflow connection name, so a generic job can call
# session.creds.cred_type(some_name_from_its_arguments) and get a
# valid cred both on the command line and in Airflow.
#
# If this seems to work, we can extend this idea out to Kubernetes
# secrets by writing a separate backend for it (and if we need to
# support both Airflow and Kubernetes secrets depending on the
# situation, let the session pick which backend we're using for
# creds based on out of band information, like how it was constructed
# or environment variables).
class BaseCreds():
    def __init__(self,
                 default_db_creds_name: Optional[str] = None,
                 default_aws_creds_name: Optional[str] = None,
                 default_gcp_creds_name: Optional[str] = None,
                 default_db_facts: Union[PleaseInfer, DBFacts] = PleaseInfer.token,
                 default_boto3_session: Union[PleaseInfer,
                                              'boto3.session.Session',
                                              None] = PleaseInfer.token,
                 default_gcp_creds: Union[PleaseInfer,
                                          'google.auth.credentials.Credentials',
                                          None] = PleaseInfer.token,
                 default_gcs_client: Union[PleaseInfer,
                                           'google.cloud.storage.Client',
                                           None] = PleaseInfer.token,
                 scratch_s3_url: Union[PleaseInfer,
                                       str,
                                       None] = PleaseInfer.token,
                 scratch_gcs_url: Union[PleaseInfer,
                                        str,
                                        None] = PleaseInfer.token) -> None:
        self._default_db_creds_name = default_db_creds_name
        self._default_aws_creds_name = default_aws_creds_name
        self._default_gcp_creds_name = default_gcp_creds_name

        self.__default_db_facts = default_db_facts
        self.__default_gcs_creds = default_gcp_creds
        self.__default_gcs_client = default_gcs_client
        self.__default_boto3_session = default_boto3_session

        self._scratch_s3_url = scratch_s3_url
        self._scratch_gcs_url = scratch_gcs_url

    def google_sheets(self, gcp_creds_name: str) -> 'google.auth.credentials.Credentials':
        scopes = _GSHEETS_SCOPES
        return self._gcp_creds(gcp_creds_name, scopes)

    def gcs(self, gcp_creds_name: str) -> 'google.auth.credentials.Credentials':
        scopes = _GCS_SCOPES
        return self._gcp_creds(gcp_creds_name, scopes)

    def _gcp_creds(self, gcp_creds_name: str,
                   scopes: Iterable[str]) -> 'google.auth.credentials.Credentials':
        raise NotImplementedError

    def db_facts(self, db_creds_name: str) -> DBFacts:
        raise NotImplementedError

    def boto3_session(self, aws_creds_name: str) -> 'boto3.session.Session':
        raise NotImplementedError

    def default_boto3_session(self) -> Optional['boto3.session.Session']:
        if self.__default_boto3_session is not PleaseInfer.token:
            return self.__default_boto3_session

        try:
            import boto3  # noqa
        except ModuleNotFoundError:
            logger.debug("boto3 not installed",
                         exc_info=True)
            return None

        if self._default_aws_creds_name is None:
            self.__default_boto3_session = boto3.session.Session()
        else:
            self.__default_boto3_session = self.boto3_session(self._default_aws_creds_name)
        return self.__default_boto3_session

    def _gcp_creds_of_last_resort(self,
                                  scopes: Iterable[str]) ->\
            Optional['google.auth.credentials.Credentials']:
        import google.auth
        credentials, project = google.auth.default(scopes=scopes)
        return credentials

    def default_gcs_creds(self) -> Optional['google.auth.credentials.Credentials']:
        if self.__default_gcs_creds is not PleaseInfer.token:
            return self.__default_gcs_creds

        try:
            import google.auth.exceptions
            if self._default_gcp_creds_name is None:
                self.__default_gcs_creds = self._gcp_creds_of_last_resort(scopes=_GCS_SCOPES)
            else:
                self.__default_gcs_creds = self.gcs(self._default_gcp_creds_name)
        except (OSError, google.auth.exceptions.DefaultCredentialsError) as e:
            # Examples:
            #   OSError: Project was not passed and could not be determined from the environment.
            #   google.auth.exceptions.DefaultCredentialsError: Could not automatically determine
            #     credentials. Please set GOOGLE_APPLICATION_CREDENTIALS or explicitly create
            #     credentials and re-run the application. For more information, please see
            #     https://cloud.google.com/docs/authentication/getting-started
            logger.warning(f"google.cloud.storage not configured: {e}")
            logger.debug("Details:", exc_info=True)
            self.__default_gcs_creds = None
        return self.__default_gcs_creds

    def default_gcs_client(self) -> Optional['google.cloud.storage.Client']:
        if self.__default_gcs_client is not PleaseInfer.token:
            return self.__default_gcs_client

        gcs_creds = self.default_gcs_creds()
        if gcs_creds is None:
            self.__default_gcs_client = None
            return self.__default_gcs_client
        try:
            import google.cloud.storage  # noqa
        except ModuleNotFoundError as e:
            logger.warning(f"google.cloud.storage not installed: {e}")
            logger.debug("Details:", exc_info=True)
            self.__default_gcs_client = None
            return self.__default_gcs_client

        try:
            other_args = {}
            # Pulling the project here from global config instead of
            # from the Google API and letting the user configure it
            # isn't great.  See
            # https://github.com/bluelabsio/records-mover/issues/119
            # for a way we can improve this.
            gcp_project = self._default_gcp_project()
            if gcp_project is not None:
                other_args['project'] = gcp_project

            self.__default_gcs_client = google.cloud.storage.Client(credentials=gcs_creds,
                                                                    **other_args)
            return self.__default_gcs_client
        except OSError as e:
            # Example:
            #   OSError: Project was not passed and could not be determined from the environment.
            logger.warning(f"google.cloud.storage not configured: {e}")
            logger.debug("Details:", exc_info=True)
            self.__default_gcs_client = None
            return self.__default_gcs_client

    def _default_gcp_project(self) -> Optional[str]:
        if 'GCP_PROJECT' in os.environ:
            return os.environ['GCP_PROJECT']
        gcp_cfg = self._config_section('gcp')
        if gcp_cfg is not None:
            default_gcp_project: Optional[str] = gcp_cfg.get('default_project')
            return default_gcp_project

        return None

    def default_db_facts(self) -> DBFacts:
        if self.__default_db_facts is not PleaseInfer.token:
            return self.__default_db_facts

        if self._default_db_creds_name is None:
            self.__default_db_facts = db_facts_from_env()
        else:
            self.__default_db_facts = self.db_facts(self._default_db_creds_name)
        return self.__default_db_facts

    def _append_aws_username_to_bucket(self,
                                       prefix: str,
                                       boto3_session: 'boto3.session.Session') -> Optional[str]:
        sts_client = boto3_session.client('sts')
        caller_identity = sts_client.get_caller_identity()
        arn = caller_identity['Arn']
        last_section_of_arn = arn.split(':')[-1]
        # Check that this is an actual user and not, say, an assumed
        # role or something else.
        if last_section_of_arn.startswith('user/'):
            username = last_section_of_arn.split('/')[-1]
            return f"{prefix}{username}/"
        else:
            logger.warning('Cannot generate S3 scratch URL with IAM username, '
                           f'as there is no username in {arn}')
            return None

    def _config_section(self, section_name: str) -> Optional[Dict[str, Any]]:
        config_result = get_config('records_mover', 'bluelabs')
        cfg = config_result.config
        if section_name in cfg:
            return cfg[section_name]
        else:
            return None

    def _infer_scratch_s3_url(self,
                              boto3_session: Optional['boto3.session.Session']) -> Optional[str]:
        if "SCRATCH_S3_URL" in os.environ:
            return os.environ["SCRATCH_S3_URL"]

        aws_cfg = self._config_section('aws')
        if aws_cfg is not None:
            s3_scratch_url: Optional[str] = aws_cfg.get('s3_scratch_url')
            if s3_scratch_url is not None:
                return s3_scratch_url
            else:
                s3_scratch_url_prefix: Optional[str] =\
                    aws_cfg.get('s3_scratch_url_appended_with_iam_username')
                if s3_scratch_url_prefix is not None:
                    if boto3_session is None:
                        logger.warning('Cannot generate S3 scratch URL with IAM username, '
                                       'as I have no IAM username')
                        return None
                    return self._append_aws_username_to_bucket(s3_scratch_url_prefix,
                                                               boto3_session)
                else:
                    logger.debug('No S3 scratch bucket config found')
                    return None
        else:
            logger.debug('No config ini file found')
            return None

    def default_scratch_s3_url(self) -> Optional[str]:
        if self._scratch_s3_url is PleaseInfer.token:
            self._scratch_s3_url = self._infer_scratch_s3_url(self.default_boto3_session())
        return self._scratch_s3_url

    def _infer_scratch_gcs_url(self) -> Optional[str]:
        if "SCRATCH_GCS_URL" in os.environ:
            return os.environ["SCRATCH_GCS_URL"]

        gcp_cfg = self._config_section('gcp')
        if gcp_cfg is not None:
            gcs_scratch_url: Optional[str] = gcp_cfg.get('gcs_scratch_url')
            if gcs_scratch_url is not None:
                return gcs_scratch_url
            else:
                logger.debug('No GCS scratch bucket config found')
                return None
        else:
            logger.debug('No config ini file found')
            return None

    def default_scratch_gcs_url(self) -> Optional[str]:
        if self._scratch_gcs_url is PleaseInfer.token:
            self._scratch_gcs_url = self._infer_scratch_gcs_url()
        return self._scratch_gcs_url



from records_mover.utils import quiet_remove
from records_mover.records.delimited import cant_handle_hint, ValidatedRecordsHints
from typing import Set, Tuple, Optional
from .types import DateOrderStyle, DateOutputStyle


def determine_date_output_style(unhandled_hints: Set[str],
                                hints: ValidatedRecordsHints,
                                fail_if_cant_handle_hint: bool) -> \
        Tuple[DateOutputStyle, Optional[DateOrderStyle]]:

    # see docs in the types module

    dateformat = hints.dateformat
    timeonlyformat = hints.timeonlyformat
    datetimeformattz = hints.datetimeformattz
    datetimeformat = hints.datetimeformat

    date_order_style: Optional[DateOrderStyle] = None

    if (dateformat == 'YYYY-MM-DD' and
       timeonlyformat in ['HH24:MI:SS', 'HH:MI:SS'] and
       datetimeformattz in ['YYYY-MM-DD HH:MI:SSOF',
                            'YYYY-MM-DD HH24:MI:SSOF'] and
       datetimeformat in ['YYYY-MM-DD HH24:MI:SS',
                          'YYYY-MM-DD HH:MI:SS']):
        date_output_style: DateOutputStyle = 'ISO'
        # date_order_style doesn't really matter, as ISO is not ambiguous
    else:
        # 'SQL', 'Postgres' and 'German' all support only alphabetic
        # timezone indicators, which aren't yet supported in the
        # records spec
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformattz', hints)

    quiet_remove(unhandled_hints, 'dateformat')
    quiet_remove(unhandled_hints, 'timeonlyformat')
    quiet_remove(unhandled_hints, 'datetimeformattz')
    quiet_remove(unhandled_hints, 'datetimeformat')

    return (date_output_style, date_order_style)



from .types import PartialRecordsHints, UntypedRecordsHints
from .validated_records_hints import ValidatedRecordsHints
import logging
from typing import Iterable, Union


logger = logging.getLogger(__name__)


def _hint_value(hints: Union[PartialRecordsHints,
                             UntypedRecordsHints,
                             ValidatedRecordsHints],
                hint_name: str) -> object:
    if isinstance(hints, ValidatedRecordsHints):
        value = getattr(hints, hint_name.replace('-', '_'))
    else:
        value = hints[hint_name]
    return value


def complain_on_unhandled_hints(fail_if_dont_understand: bool,
                                unhandled_hints: Iterable[str],
                                hints: Union[PartialRecordsHints,
                                             UntypedRecordsHints,
                                             ValidatedRecordsHints]) -> None:
    unhandled_bindings = [f"{k}={_hint_value(hints, k)}" for k in unhandled_hints]
    unhandled_bindings_str = ", ".join(unhandled_bindings)
    if len(unhandled_bindings) > 0:
        if fail_if_dont_understand:
            err = "Implement these records_format hints or try again with " +\
                f"fail_if_dont_understand=False': {unhandled_bindings_str}"
            raise NotImplementedError(err)
        else:
            logger.warning(f"Did not understand these hints: {unhandled_bindings_str}")


def cant_handle_hint(fail_if_cant_handle_hint: bool,
                     hint_name: str,
                     hints: Union[PartialRecordsHints,
                                  UntypedRecordsHints,
                                  ValidatedRecordsHints]) -> None:
    value = _hint_value(hints, hint_name)
    if not fail_if_cant_handle_hint:
        logger.warning("Ignoring hint {hint_name} = {hint_value}"
                       .format(hint_name=hint_name,
                               hint_value=repr(value)))
    else:
        raise NotImplementedError(f"Implement hint {hint_name}="
                                  f"{repr(value)} " +
                                  "or try again with fail_if_cant_handle_hint=False")



from typing import Optional

# An arbitrary 4 mb csv I looked at ran around 100,000 lines.
# Assuming we want to limit our memory usage to, say, 400MB of memory,
# let's limit inference to 1,000,000 lines.
DEFAULT_MAX_SAMPLE_SIZE = 1000000


class ProcessingInstructions:
    def __init__(self,
                 fail_if_dont_understand: bool=True,
                 fail_if_cant_handle_hint: bool=True,
                 fail_if_row_invalid: bool=True,
                 max_inference_rows: Optional[int]=DEFAULT_MAX_SAMPLE_SIZE,
                 max_failure_rows: Optional[int]=None) -> None:
        """Directives on how to handle different situations when processing
        records.  Note that not all vendor mechanisms support this
        level of configurability; when choosing between optimizing for
        fast transfer and ability to comply, Records Mover will favor
        fast transfer.

        :param fail_if_dont_understand: If True, and a part of the RecordsFormat is not understood
           while processing, then immediately fail and raise an exception.  Otherwise, ignore the
           misunderstood instruction (e.g., ignore the hint, assume default variant, etc etc)

        :param fail_if_cant_handle_hint: If True, and for whatever reason (e.g., limited options in
           whatever library/tool/database is being used) a certain hint can't be handled as
           specified, raise an exception.  Otherwise, ignore the hint and use
           implementation-specific different behavior.

        :param fail_if_row_invalid: If True, and a particular row of data in the records file
           cannot be understood by the library, raise an exception.  Otherwise, ignore the row and
           continue and try to load other rows.

        :param max_failure_rows: Sets a tolerance level for number of rows of data in the records
           file that cannot be understood by the library that should be ignored. After reaching
           level, raise an exception.

        :param max_inference_rows: If the schema is not provided and we need it (e.g., we're to
           load the records into a database and there's no existing table), we'll figure it out
           through 'type inference' - looking at a bunch of examples of data and building a
           specific schema that can load those rows.  This can take some time, so this parameter
           controls the maximum number of rows we'll look at.  Higher values will be more likely to
           result in a schema that can be loaded into, but will take longer to load.  If set to
           None, the entire file will be processed.
        """

        self.fail_if_dont_understand = fail_if_dont_understand
        self.fail_if_cant_handle_hint = fail_if_cant_handle_hint
        self.fail_if_row_invalid = fail_if_row_invalid
        self.max_failure_rows = max_failure_rows
        self.max_inference_rows = max_inference_rows



from typing_inspect import get_args
from typing import Mapping, List
from typing_extensions import Literal
# TypedDict isn't mypy specific, but typing_inspect currently doesn't
# support typing_extensions.TypedDict.
#
# https://github.com/ilevkivskyi/typing_inspect/issues/50
from mypy_extensions import TypedDict

HintEncoding = Literal["UTF8", "UTF16", "UTF16LE", "UTF16BE",
                       "UTF16BOM", "UTF8BOM", "LATIN1", "CP1252"]

HintQuoting = Literal["all", "minimal", "nonnumeric", None]

HintEscape = Literal["\\", None]

HintCompression = Literal['GZIP', 'BZIP', 'LZO', None]

# The trick here works on Literal[True, False] but not on bool:
#
# https://github.com/python/mypy/issues/6366#issuecomment-560369716
HintHeaderRow = Literal[True, False]

HintDoublequote = Literal[True, False]

HintDateFormat = str

HintTimeOnlyFormat = str

HintDateTimeFormatTz = str

HintDateTimeFormat = str

HintFieldDelimiter = str

HintRecordTerminator = str

HintQuoteChar = str

HintName = Literal["header-row",
                   "field-delimiter",
                   "compression",
                   "record-terminator",
                   "quoting",
                   "quotechar",
                   "doublequote",
                   "escape",
                   "encoding",
                   "dateformat",
                   "timeonlyformat",
                   "datetimeformattz",
                   "datetimeformat"]

HINT_NAMES: List[HintName] = list(get_args(HintName))

PartialRecordsHints = TypedDict('PartialRecordsHints',
                                {
                                    'quoting': HintQuoting,
                                    'header-row': HintHeaderRow,
                                    'field-delimiter': HintFieldDelimiter,
                                    'encoding': HintEncoding,
                                    'escape': HintEscape,
                                    'compression': HintCompression,
                                    'record-terminator': HintRecordTerminator,
                                    'quotechar': HintQuoteChar,
                                    'doublequote': HintDoublequote,
                                    'dateformat': HintDateFormat,
                                    'timeonlyformat': HintTimeOnlyFormat,
                                    'datetimeformattz': HintDateTimeFormatTz,
                                    'datetimeformat': HintDateTimeFormat,
                                },
                                total=False)
UntypedRecordsHints = Mapping[str, object]



from urllib.parse import urlparse
from .types import HintCompression
import os


def sniff_compression_from_url(url: str) -> HintCompression:
    "Returns compression hint for a given file extension"
    urlobj = urlparse(url)
    pathname = urlobj.path
    if pathname:
        splitpathname = os.path.splitext(pathname)
        if len(splitpathname) != 2:
            return None
        ext = splitpathname[1]
        if ext:
            if ext.lower() == '.gz':
                return 'GZIP'
            elif ext.lower() == '.bz2':
                return "BZIP"
            elif ext.lower() == '.lzo':
                return "LZO"
    return None



from typing import Optional
from sqlalchemy.engine import Connection
from records_mover.db.quoting import quote_schema_and_table
from records_mover.records.existing_table_handling import ExistingTableHandling
from records_mover.db import DBDriver
from records_mover.records.table import TargetTableDetails
import logging

logger = logging.getLogger(__name__)


class TablePrep:
    def __init__(self, target_table_details: TargetTableDetails) -> None:
        self.tbl = target_table_details

    def add_permissions(self, conn: Connection, driver: DBDriver) -> None:
        schema_and_table: str = quote_schema_and_table(driver.db,
                                                       self.tbl.schema_name,
                                                       self.tbl.table_name)
        if self.tbl.add_group_perms_for is not None:
            logger.info(f"Adding permissions for {schema_and_table} "
                        f"to group {self.tbl.add_group_perms_for}")
            driver.set_grant_permissions_for_groups(self.tbl.schema_name,
                                                    self.tbl.table_name,
                                                    self.tbl.add_group_perms_for,
                                                    conn)
        if self.tbl.add_user_perms_for is not None:
            logger.info(f"Adding permissions for {schema_and_table} "
                        f"to {self.tbl.add_user_perms_for}")
            driver.\
                set_grant_permissions_for_users(self.tbl.schema_name,
                                                self.tbl.table_name,
                                                self.tbl.add_user_perms_for,
                                                conn)

    def create_table(self, schema_sql: str, conn: Connection, driver: DBDriver) -> None:
        logger.info('Creating table...')
        conn.execute(schema_sql)
        logger.info(f"Just ran {schema_sql}")
        self.add_permissions(conn, driver)
        logger.info("Table prepped")

    def prep_table_for_load(self,
                            schema_sql: str,
                            existing_table_handling: ExistingTableHandling,
                            driver: DBDriver) -> None:
        logger.info("Looking for existing table..")
        db = driver.db

        if driver.has_table(table=self.tbl.table_name, schema=self.tbl.schema_name):
            logger.info("Table already exists.")
            how_to_prep = existing_table_handling
            schema_and_table: str = quote_schema_and_table(db,
                                                           self.tbl.schema_name,
                                                           self.tbl.table_name)
            if (how_to_prep == ExistingTableHandling.TRUNCATE_AND_OVERWRITE):
                logger.info("Truncating...")
                db.execute(f"TRUNCATE TABLE {schema_and_table}")
                logger.info("Truncated.")
            elif (how_to_prep == ExistingTableHandling.DELETE_AND_OVERWRITE):
                logger.info("Deleting rows...")
                db.execute(f"DELETE FROM {schema_and_table} WHERE true")
                logger.info("Deleted")
            elif (how_to_prep == ExistingTableHandling.DROP_AND_RECREATE):
                with db.engine.connect() as conn:
                    with conn.begin():
                        logger.info("Dropping and recreating...")
                        drop_table_sql = f"DROP TABLE {schema_and_table}"
                        conn.execute(drop_table_sql)
                        logger.info(f"Just ran {drop_table_sql}")
                        self.create_table(schema_sql, conn, driver)
            elif (how_to_prep == ExistingTableHandling.APPEND):
                logger.info("Appending rows...")
            else:
                raise ValueError(f"Don't know how to handle {how_to_prep}")
        else:
            with db.engine.connect() as conn:
                with conn.begin():
                    self.create_table(schema_sql, conn, driver)

    def prep(self,
             schema_sql: str,
             driver: DBDriver,
             existing_table_handling: Optional[ExistingTableHandling] = None) -> None:
        if existing_table_handling is None:
            existing_table_handling = self.tbl.existing_table_handling
        self.prep_table_for_load(schema_sql=schema_sql,
                                 existing_table_handling=existing_table_handling,
                                 driver=driver)



from records_mover.utils import quiet_remove
from records_mover.records.delimited import cant_handle_hint, ValidatedRecordsHints
from typing import Set
from .types import PostgresCopyOptions


# https://www.postgresql.org/docs/9.3/multibyte.html
postgres_encoding_names = {
    'UTF8': 'UTF8',
    # UTF-16 not supported
    # https://www.postgresql.org/message-id/20051028061108.GB30916%40london087.server4you.de
    'LATIN1': 'LATIN1',
    'CP1252': 'WIN1252',
}


def postgres_copy_options_common(unhandled_hints: Set[str],
                                 hints: ValidatedRecordsHints,
                                 fail_if_cant_handle_hint: bool,
                                 original_postgres_options: PostgresCopyOptions) ->\
        PostgresCopyOptions:
    postgres_options = original_postgres_options.copy()

    # ENCODING
    #
    #  Specifies that the file is encoded in the encoding_name. If
    #  this option is omitted, the current client encoding is
    #  used. See the Notes below for more details.

    encoding_hint = hints.encoding
    if encoding_hint in postgres_encoding_names:
        postgres_options['encoding'] = postgres_encoding_names[encoding_hint]
        quiet_remove(unhandled_hints, 'encoding')
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'encoding', hints)

    # FORCE_NOT_NULL
    #
    #  Do not match the specified columns' values against the null
    #  string. In the default case where the null string is empty,
    #  this means that empty values will be read as zero-length
    #  strings rather than nulls, even when they are not quoted. This
    #  option is allowed only in COPY FROM, and only when using CSV
    #  format.
    #
    # HEADER
    #
    #  Specifies that the file contains a header line with the names
    #  of each column in the file. On output, the first line contains
    #  the column names from the table, and on input, the first line
    #  is ignored. This option is allowed only when using CSV format.
    #
    quiet_remove(unhandled_hints, 'header-row')
    postgres_options['header'] = hints.header_row

    # OIDS
    #
    #  Specifies copying the OID for each row. (An error is raised if
    #  OIDS is specified for a table that does not have OIDs, or in
    #  the case of copying a query.)
    #

    # DELIMITER
    #
    #  Specifies the character that separates columns within each row
    #  (line) of the file. The default is a tab character in text
    #  format, a comma in CSV format. This must be a single one-byte
    #  character. This option is not allowed when using binary format.
    #
    postgres_options['delimiter'] = hints.field_delimiter
    quiet_remove(unhandled_hints, 'field-delimiter')

    return postgres_options



from ...utils.json_schema import method_signature_to_json_schema, JsonParameter, JsonSchemaDocument
from ..existing_table_handling import ExistingTableHandling
from records_mover.records.delimited.hints import Hints
from records_mover.records.records_types import RECORDS_FORMAT_TYPES
from typing import Any, Dict, List, Callable
from ...mover_types import JsonSchema


HINT_PARAMETERS = [
    JsonParameter(hint_enum.value.hint_name,
                  hint_enum.value.json_schema_document(),
                  optional=True)
    for hint_enum in list(Hints)
]


def method_to_json_schema(method: Callable[..., Any]) -> JsonSchema:
    special_handling: Dict[str, List[JsonParameter]] = {
        'google_cloud_creds': [JsonParameter('gcp_creds_name', JsonSchemaDocument('string'))],
        'db_engine': [JsonParameter('db_name', JsonSchemaDocument('string'))],
        'records_format': ([JsonParameter('variant',
                                          JsonSchemaDocument('string',
                                                             description="Records format variant - "
                                                             "valid for 'delimited' "
                                                             "records format type"),
                                          optional=True),
                            JsonParameter('format',
                                          JsonSchemaDocument('string',
                                                             enum=RECORDS_FORMAT_TYPES,
                                                             description="Records format type.  "
                                                             "Note that 'delimited' includes "
                                                             "CSV/TSV/etc."),
                                          optional=True,)] +
                           HINT_PARAMETERS),
        'initial_hints': HINT_PARAMETERS,
        'existing_table_handling':
        [JsonParameter('existing_table',
                       JsonSchemaDocument('string',
                                          enum=[k.lower()
                                                for k in ExistingTableHandling.__members__],
                                          default='delete_and_overwrite'),
                       optional=True)],
    }

    parameters_to_ignore = [
        'self',
        # maybe in the future we can have people point us to a JSON
        # file for hints...
        'hints',
        # ...and maybe this particular way of passing in dicts:
        'add_user_perms_for',
        'add_group_perms_for',
        # oh yeah, and a way to pass a filename in for those that have
        # hand-crafted a records schema file yet don't have a full
        # records directory...
        'records_schema'
    ]

    return method_signature_to_json_schema(method,
                                           special_handling=special_handling,
                                           parameters_to_ignore=parameters_to_ignore)



import logging
import json
from typing import List, Dict, Mapping, IO, Any, TYPE_CHECKING
from ..field import RecordsSchemaField
from ...records_format import BaseRecordsFormat
from ...processing_instructions import ProcessingInstructions
from .known_representation import RecordsSchemaKnownRepresentation
from ..errors import UnsupportedSchemaError
if TYPE_CHECKING:
    from pandas import DataFrame

    from ....db import DBDriver  # noqa
    from typing_extensions import Literal

    from mypy_extensions import TypedDict

    from ..field import FieldDict
    from ..field.field_types import FieldType  # noqa
    from .known_representation import KnownRepresentationDict

    class MandatoryRecordsSchemaDict(TypedDict):
        schema: Literal['bltypes/v1']
        fields: Dict[str, FieldDict]

    class RecordsSchemaDict(MandatoryRecordsSchemaDict, total=False):
        known_representations: Dict[str, KnownRepresentationDict]


logger = logging.getLogger(__name__)


class RecordsSchema:
    """This class records whatever type information we have available at
    the time of capture of records data so that future readers of the
    data can do a minimum of type inference on the remaining data for
    whatever the target system is.

    See the `Records Schema spec
    <https://github.com/bluelabsio/records-mover/blob/master/docs/schema/SCHEMAv1.md>`_
    for more details.
    """

    def __init__(self,
                 fields: List[RecordsSchemaField],
                 known_representations: Mapping[str, RecordsSchemaKnownRepresentation]) -> None:
        """
        :param fields: Ordered list of which fields are included in this schema.
        :param known_representations: Detailed information about how each field is intended to be
           represented in certain systems.  The key is the name of the known representation type
           (e.g., "origin" or some other short nickname for the type if it is not the origin - e.g.,
           "redshift"). The value is an object with a type field containing the type of source
           system (e.g., sql/redshift) and other, type-specific fields that may be useful for
           reconstructing the schema in the target system.
        """
        self.fields = fields
        self.known_representations = known_representations

    def __str__(self) -> str:
        types = {field.name: field.field_type for field in self.fields}
        return f"RecordsSchema(types={types})"

    @staticmethod
    def from_db_table(schema_name: str, table_name: str, driver: 'DBDriver') -> 'RecordsSchema':
        from .sqlalchemy import schema_from_db_table
        return schema_from_db_table(schema_name=schema_name,
                                    table_name=table_name,
                                    driver=driver)

    def to_data(self) -> 'RecordsSchemaDict':
        data: 'RecordsSchemaDict' = {
            'schema': 'bltypes/v1',
            'fields': {},
        }
        i = 1
        for field in self.fields:
            data['fields'][field.name] = field.to_data()
            data['fields'][field.name]['index'] = i
            i = i + 1
        if self.known_representations:
            data['known_representations'] = {
                system: known_representation.to_data()
                for system, known_representation
                in self.known_representations.items()
            }
        return data

    def to_json(self) -> str:
        return json.dumps(self.to_data())

    @staticmethod
    def from_data(data: 'RecordsSchemaDict') -> 'RecordsSchema':
        """Create a RecordsSchema object from a Python dictionary serialized form.

        :param data: Python dictionary containing the serialized data described in the `Records
           Schema spec
           <https://github.com/bluelabsio/records-mover/blob/master/docs/schema/SCHEMAv1.md>`_
        :return: RecordsSchema object suitable for passing to Records Mover source/target factory
           methods.
        """
        schema_ver = data["schema"]
        if schema_ver != "bltypes/v1":
            raise UnsupportedSchemaError(
                f"unsupported record specification schema {schema_ver}")
        known_representations = {}
        for name, known_rep_data in data.get("known_representations", {}).items():
            known_representation = RecordsSchemaKnownRepresentation.from_data(known_rep_data)
            if known_representation is not None:
                known_representations[name] = known_representation
        return RecordsSchema(
            fields=[
                RecordsSchemaField.from_data(name, field_data)
                for name, field_data
                in data.get("fields", {}).items()
            ],
            known_representations=known_representations)

    @staticmethod
    def from_json(json_str: str) -> 'RecordsSchema':
        data = json.loads(json_str)
        return RecordsSchema.from_data(data)

    def to_schema_sql(self,
                      driver: 'DBDriver',
                      schema_name: str,
                      table_name: str) -> str:
        from .sqlalchemy import schema_to_schema_sql
        return schema_to_schema_sql(records_schema=self,
                                    driver=driver,
                                    schema_name=schema_name,
                                    table_name=table_name)

    @staticmethod
    def from_fileobjs(fileobjs: List[IO[bytes]],
                      records_format: BaseRecordsFormat,
                      processing_instructions: ProcessingInstructions) -> 'RecordsSchema':
        """
        Sniffs
        """
        from records_mover.records.delimited import stream_csv
        from records_mover.pandas import purge_unnamed_unused_columns

        if len(fileobjs) != 1:
            # https://github.com/bluelabsio/records-mover/issues/84
            raise NotImplementedError('Cannot currently sniff schema from multiple '
                                      'files--please provide explicit schema JSON')
        fileobj = fileobjs[0]
        if not fileobj.seekable():
            raise NotImplementedError('Cannot currently sniff schema from a pure stream--'
                                      'please save file to disk and load from there or '
                                      'provide explicit schema JSON')
        with stream_csv(fileobj, records_format.hints) as reader:  # type: ignore
            # Parse schema from sample df

            sample_row_count = processing_instructions.max_inference_rows
            if sample_row_count is not None:
                df = reader.get_chunk(sample_row_count)
            else:
                df = reader.read()

            fileobj.seek(0)

            df = purge_unnamed_unused_columns(df)
            schema = RecordsSchema.from_dataframe(df, processing_instructions,
                                                  include_index=False)

            schema = schema.refine_from_dataframe(df,
                                                  processing_instructions=processing_instructions)
            return schema

    def refine_from_dataframe(self,
                              df: 'DataFrame',
                              processing_instructions:
                              ProcessingInstructions = ProcessingInstructions()) -> 'RecordsSchema':
        """
        Adjust records schema based on facts found from a dataframe.
        """
        from .pandas import refine_schema_from_dataframe
        return refine_schema_from_dataframe(records_schema=self,
                                            df=df,
                                            processing_instructions=processing_instructions)

    def cast_dataframe_types(self,
                             df: 'DataFrame') -> 'DataFrame':
        """
        Returns a new dataframe with types that match what we know from this records schema.
        """
        fields = {field.name: field for field in self.fields}
        return df.apply(lambda series: fields[series.name].cast_series_type(series))

    def assign_dataframe_names(self,
                               include_index: bool,
                               df: 'DataFrame') -> 'DataFrame':
        """Returns a new dataframe with index/series names that match what we
        know from this records schema.  Useful when we've created a
        dataframe from a file where the headers don't match the exact
        names or when the names aren't provided
        """
        df_names = [label for label in df]
        field_names = [field.name for field in self.fields]
        index_mapping = None
        if include_index:
            index_name = field_names[0]
            # RecordsMover supports only single indexes for the moment.
            #
            # https://github.com/bluelabsio/records-mover/issues/92
            index_mapping = {0: index_name}

            # remove first name, which applies to index
            field_names = field_names[1:]
        if len(df_names) != len(field_names):
            raise SyntaxError("Number of DataFrame names didn't match number of schema names!"
                              f"DataFrame: {df_names}.  Schema names: {field_names}")
        name_mapping = dict(zip(df_names, field_names))
        df = df.rename(columns=name_mapping,
                       copy=False,
                       errors='raise')
        if include_index:
            return df.rename(index=index_mapping,
                             copy=False,
                             errors='raise')
        else:
            return df

    @staticmethod
    def from_dataframe(df: 'DataFrame',
                       processing_instructions: ProcessingInstructions,
                       include_index: bool) -> 'RecordsSchema':
        """Create a RecordsSchema object representing a Pandas dataframe.

        :param df: Pandas dataframe that should be analyzed to determine schema information.
        :param processing_instructions: Instructions used during creation of the records schema,
           including how much data to analyze to infer this schema.  This is of type
           :class:`records_mover.records.ProcessingInstructions`
        :param include_index: If true, the Pandas dataframe index column will be included in the
           move.

        :return: RecordsSchema object suitable for passing to Records Mover source/target factory
           methods.
        """
        from .pandas import schema_from_dataframe
        return schema_from_dataframe(df=df,
                                     processing_instructions=processing_instructions,
                                     include_index=include_index)

    def to_empty_dataframe(self) -> 'DataFrame':
        from pandas import DataFrame
        df_data: Dict[str, List[Any]] = {field.name: [] for field in self.fields}
        raw_df = DataFrame(df_data)
        return self.cast_dataframe_types(raw_df)

    def cast_field_types(self, typecasts: Dict['FieldType', 'FieldType']) -> 'RecordsSchema':
        return RecordsSchema(fields=[field.cast(typecasts.get(field.field_type,
                                                              field.field_type))
                                     for field in self.fields],
                             known_representations=self.known_representations)



from ...utils import quiet_remove
from ...records.delimited import cant_handle_hint
from ...records.unload_plan import RecordsUnloadPlan
from ...records.records_format import DelimitedRecordsFormat
from typing import Set, Dict, Any


def vertica_export_options(unhandled_hints: Set[str],
                           unload_plan: RecordsUnloadPlan) -> Dict[str, Any]:
    if not isinstance(unload_plan.records_format, DelimitedRecordsFormat):
        raise NotImplementedError("Not currently able to export "
                                  f"{unload_plan.records_format.format_type}")

    hints = unload_plan.records_format.hints
    fail_if_cant_handle_hint = unload_plan.processing_instructions.fail_if_cant_handle_hint

    vertica_options = {}

    vertica_options['to_charset'] = hints['encoding']
    quiet_remove(unhandled_hints, 'encoding')

    if hints['compression'] is not None:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)
    quiet_remove(unhandled_hints, 'compression')

    if hints['quoting'] is not None:
        cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
    quiet_remove(unhandled_hints, 'quoting')
    quiet_remove(unhandled_hints, 'doublequote')
    quiet_remove(unhandled_hints, 'quotechar')

    # Vertica does no escaping with S3EXPORT:
    #   https://my.vertica.com/docs/8.1.x/HTML/index.htm#Authoring/
    #     SQLReferenceManual/Functions/VerticaFunctions/s3export.htm
    if hints['escape'] is not None:
        cant_handle_hint(fail_if_cant_handle_hint, 'escape', hints)
    quiet_remove(unhandled_hints, 'escape')

    if hints['header-row'] is not False:
        cant_handle_hint(fail_if_cant_handle_hint, 'header-row', hints)
    quiet_remove(unhandled_hints, 'header-row')

    if hints['dateformat'] != 'YYYY-MM-DD':
        cant_handle_hint(fail_if_cant_handle_hint, 'dateformat', hints)
    quiet_remove(unhandled_hints, 'dateformat')

    if hints['datetimeformattz'] not in ['YYYY-MM-DD HH:MI:SSOF',
                                         'YYYY-MM-DD HH24:MI:SSOF']:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformattz', hints)
    quiet_remove(unhandled_hints, 'datetimeformattz')

    if hints['datetimeformat'] not in ['YYYY-MM-DD HH:MI:SS',
                                       'YYYY-MM-DD HH24:MI:SS']:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', hints)
    quiet_remove(unhandled_hints, 'datetimeformat')

    if hints['timeonlyformat'] not in ['HH24:MI:SS',
                                       'HH:MI:SS']:
        cant_handle_hint(fail_if_cant_handle_hint, 'timeonlyformat', hints)
    quiet_remove(unhandled_hints, 'timeonlyformat')

    vertica_options['delimiter'] = hints['field-delimiter']
    quiet_remove(unhandled_hints, 'field-delimiter')

    vertica_options['record_terminator'] = hints['record-terminator']
    quiet_remove(unhandled_hints, 'record-terminator')

    return vertica_options



from abc import ABCMeta, abstractmethod
import json
from collections import OrderedDict
import logging
import re
from typing import Optional, Dict, Union, Any, cast, TYPE_CHECKING
if TYPE_CHECKING:
    from sqlalchemy.engine.interfaces import Dialect
    from sqlalchemy import Column
    from typing_extensions import Literal
    from mypy_extensions import TypedDict
    import pandas

    class FieldRepresentationDict(TypedDict):
        rep_type: str

    class MandatorySqlFieldRepresentationDict(FieldRepresentationDict):
        col_ddl: str

    class SqlFieldRepresentationDict(MandatorySqlFieldRepresentationDict, total=False):
        col_type: str
        col_modifiers: str

    class CorePandasFieldRepresentationDict(FieldRepresentationDict):
        pd_df_dtype: Dict[str, Any]
        pd_df_coltype: Union[Literal['index'], Literal['series']]

    class PandasFieldRepresentationDict(CorePandasFieldRepresentationDict, total=False):
        pd_df_ftype: str


logger = logging.getLogger(__name__)


class RecordsSchemaFieldRepresentation(metaclass=ABCMeta):
    @staticmethod
    def from_data(data: 'FieldRepresentationDict') -> Optional['RecordsSchemaFieldRepresentation']:
        if data['rep_type'].startswith('sql/'):
            sql_data = cast('SqlFieldRepresentationDict', data)
            return RecordsSchemaSqlFieldRepresentation(col_ddl=sql_data['col_ddl'],
                                                       rep_type=sql_data['rep_type'],
                                                       col_type=sql_data.get('col_type'),
                                                       col_modifiers=sql_data.get('col_modifiers'))
        elif data['rep_type'] == 'dataframe/pandas':
            pandas_data = cast('PandasFieldRepresentationDict', data)
            return RecordsSchemaPandasFieldRepresentation(
                pd_df_dtype=pandas_data['pd_df_dtype'],
                pd_df_ftype=pandas_data.get('pd_df_ftype'),
                pd_df_coltype=pandas_data['pd_df_coltype'])
        else:
            logger.warning(f"I don't understand type {data['rep_type']}--"
                           "consider upgrading records-mover")
            return None

    @staticmethod
    def from_series(series: 'pandas.Series') -> 'RecordsSchemaFieldRepresentation':
        import pandas

        dtype_json_str = pandas.io.json.dumps(series.dtype)
        dtype_numpy_rep = json.loads(dtype_json_str)
        ftype = getattr(series, 'ftype', None)
        return RecordsSchemaPandasFieldRepresentation(pd_df_coltype='series',
                                                      pd_df_dtype=dtype_numpy_rep,
                                                      pd_df_ftype=ftype)

    @staticmethod
    def from_index(index: 'pandas.Index') -> 'RecordsSchemaFieldRepresentation':
        dtype_json_str = pandas.io.json.dumps(index.dtype)
        dtype_numpy_rep = json.loads(dtype_json_str)
        return RecordsSchemaPandasFieldRepresentation(pd_df_coltype='index',
                                                      pd_df_dtype=dtype_numpy_rep,
                                                      pd_df_ftype=index.ftype)

    @staticmethod
    def from_sqlalchemy_column(column: 'Column', dialect: 'Dialect',
                               rep_type: str) ->\
            'RecordsSchemaFieldRepresentation':
        from sqlalchemy.schema import CreateColumn

        statement = CreateColumn(column)
        col_ddl_compiler = statement.compile(dialect=dialect)
        full_col_ddl = str(col_ddl_compiler)

        #
        # Parse single string output from SQLAlchemy into what records
        # schema wants.  Sadly there's no universal way to get this
        # out of the SQLAlchemy drivers themselves - example code from
        # the Redshift driver:
        # https://github.com/sqlalchemy-redshift/sqlalchemy-redshift/blob/6885b059c2d23423ae763678a46bbcdd1b3db433/sqlalchemy_redshift/dialect.py#L345
        #
        # These regexps are in order from most specfic to least so
        # that we can match early on the special cases without them
        # being caught by the more general ones.  If this code is not
        # working right for your case, please see the unit test in
        # test_field_representation and enhance it first to cover your
        # newly discovered oddity.
        #

        regexps: Dict[str, str] = OrderedDict()
        # First, cover cases where the types themselves have spaces in them:
        # e.g., DOUBLE PRECISION
        regexps['two_word_type_1'] = r'^([\S]+) (DOUBLE PRECISION)$'
        # e.g,, DOUBLE PRECISION SORTKEY
        regexps['two_word_type_1_with_col_modifiers'] = r'^([\S]+) (DOUBLE PRECISION) (.*)$'
        # e.g., TIMESTAMP WITH TIME ZONE
        regexps['two_word_type_2'] = r'^([\S]+) (TIMESTAMP WITH TIME ZONE)$'
        # e.g., TIMESTAMP WITH TIME ZONE ENCODE lzo
        regexps['two_word_type_2_with_col_modifiers'] = r'^([\S]+) (TIMESTAMP WITH TIME ZONE) (.*)$'
        # e.g., TIMESTAMP WITHOUT TIME ZONE
        regexps['two_word_type_3'] = r'^([\S]+) (TIMESTAMP WITHOUT TIME ZONE)$'
        # e.g., TIMESTAMP WITH TIME ZONE ENCODE lzo
        regexps['two_word_type_3_with_col_modifiers'] =\
            r'^([\S]+) (TIMESTAMP WITHOUT TIME ZONE) (.*)$'
        # e.g., NUMERIC(2, 1)
        regexps['parenthetical_col_type'] = r'^([\S]+) ([\S]+\(.*\))$'
        # e.g., NUMERIC(2, 1) ENCODE lzo
        regexps['parenthetical_col_type_with_col_modifiers'] = r'^([\S]+) ([\S]+\(.*\)) (.*)$'
        # e.g., REAL
        regexps['boring_col_type'] = r'^([\S]+) ([\S]+)$'
        # e.g., REAL ENCODE lzo
        regexps['boring_col_type_with_col_modifiers'] = r'^([\S]+) ([\S]+) (.*)$'

        for regexp in regexps.values():
            match = re.match(regexp, full_col_ddl)
            if match is not None:
                statement_words = list(match.groups())
                break
        else:
            raise SyntaxError(f"Could not understand {full_col_ddl}")
        col_ddl = ' '.join(statement_words[1:])
        col_type = statement_words[1]
        col_modifiers: Optional[str] = ' '.join(statement_words[2:])
        if col_modifiers == '':
            col_modifiers = None
        return RecordsSchemaSqlFieldRepresentation(col_ddl=col_ddl,
                                                   rep_type=rep_type,
                                                   col_type=col_type,
                                                   col_modifiers=col_modifiers)

    @abstractmethod
    def to_data(self) -> 'FieldRepresentationDict':
        raise NotImplementedError


class RecordsSchemaSqlFieldRepresentation(RecordsSchemaFieldRepresentation):
    def __init__(self,
                 col_ddl: str,
                 rep_type: str,
                 col_type: Optional[str],
                 col_modifiers: Optional[str]):
        self.rep_type = rep_type
        self.col_ddl = col_ddl
        self.col_type = col_type
        self.col_modifiers = col_modifiers

    def to_data(self) -> 'SqlFieldRepresentationDict':
        out: 'SqlFieldRepresentationDict' = {
            'rep_type': self.rep_type,
            'col_ddl': self.col_ddl
        }
        if self.col_type is not None:
            out['col_type'] = self.col_type
        if self.col_modifiers is not None:
            out['col_modifiers'] = self.col_modifiers
        return out


class RecordsSchemaPandasFieldRepresentation(RecordsSchemaFieldRepresentation):
    def __init__(self,
                 pd_df_dtype: Dict[str, Any],
                 pd_df_ftype: Optional[str],
                 pd_df_coltype: Union['Literal["series"]', 'Literal["index"]']):
        self.pd_df_dtype = pd_df_dtype
        self.pd_df_ftype = pd_df_ftype
        self.pd_df_coltype = pd_df_coltype

    def to_data(self) -> 'PandasFieldRepresentationDict':
        out: 'PandasFieldRepresentationDict' = {
            'rep_type': 'dataframe/pandas',
            'pd_df_dtype': self.pd_df_dtype,
            'pd_df_coltype': self.pd_df_coltype,
        }
        if self.pd_df_ftype is not None:
            out['pd_df_ftype'] = self.pd_df_ftype
        return out



__all__ = [
    'PartialRecordsHints',
    'UntypedRecordsHints',
    'RecordsFormatType',
    'RecordsSchema',
    'RecordsFormat',
    'DelimitedVariant',
    'DelimitedRecordsFormat',
    'AvroRecordsFormat',
    'ParquetRecordsFormat',
    'ProcessingInstructions',
    'ExistingTableHandling',
    'RecordsFolderNonEmptyException',
    'RecordsException',
    'Records',
    'move',
    'MoveResult'
]

from .delimited import UntypedRecordsHints, PartialRecordsHints
from .records_types import RecordsFormatType, DelimitedVariant
from .schema import RecordsSchema
from .mover import move
from .records_format import (
    RecordsFormat, DelimitedRecordsFormat, ParquetRecordsFormat, AvroRecordsFormat
)
from .processing_instructions import ProcessingInstructions
from .existing_table_handling import ExistingTableHandling
from .records import Records
from .errors import RecordsException, RecordsFolderNonEmptyException
from .results import MoveResult



from typing import NamedTuple, Optional, Mapping


class MoveResult(NamedTuple):
    """Represents the result of a move() operation between a records
    source and a records target.

    Note that move_count and output_urls may be empty depending on the
    nature of the move - e.g., a move to a database doesn't currently
    map to a formal URL, and whole-file based moves do not currently
    count the number of records being moved.
    """
    move_count: Optional[int]
    "Number of rows moved (Optional[int])"

    output_urls: Optional[Mapping[str, str]]
    """A dictionary of short string aliases mapping to URLs of the
    resulting data (Optional[Mapping[str, str]])"""



from ..url.base import BaseDirectoryUrl
from typing import Optional
import logging

logger = logging.getLogger(__name__)


class RecordsSchemaJsonFile:
    def __init__(self, records_loc: BaseDirectoryUrl) -> None:
        self.schema_loc = records_loc.file_in_this_directory('_schema.json')

    def save_schema_json(self, json: str) -> None:
        logger.info(f"Putting into {self.schema_loc.url}")
        self.schema_loc.store_string(json)

    def load_schema_json(self) -> Optional[str]:
        try:
            return self.schema_loc.string_contents()
        except FileNotFoundError:
            logger.debug('No schema JSON found', exc_info=True, stack_info=True)
            return None



"""Create and run jobs to convert between different sources and targets"""
import inspect
from typing import Any, Dict, List, Callable
from records_mover import Session
from ..records_format import DelimitedRecordsFormat, ParquetRecordsFormat, BaseRecordsFormat
from ..existing_table_handling import ExistingTableHandling
from ..delimited import PartialRecordsHints
from records_mover.records.delimited.types import HINT_NAMES
from ...mover_types import JobConfig


class ConfigToArgs:
    def __init__(self,
                 config: JobConfig,
                 method: Callable[..., Any],
                 session: Session):
        self.config = config
        self.method = method
        self.session = session

    def all_arg_names(self, fn: Callable[..., Any]) -> List[str]:
        return [param_name
                for param_name, parameter in inspect.signature(fn).parameters.items()]

    def non_default_arg_names(self, fn: Callable[..., Any]) -> List[str]:
        return [param_name
                for param_name, parameter in inspect.signature(fn).parameters.items()
                if parameter.default == inspect.Parameter.empty]

    def add_hint_parameter(self, kwargs: Dict[str, Any], param_name: str) -> None:
        if 'records_format' in kwargs:
            existing_records_format = kwargs['records_format']
            if not isinstance(existing_records_format, DelimitedRecordsFormat):
                raise NotImplementedError('Hints are not compatible '
                                          'with records format '
                                          f'{existing_records_format.format_type}')
            # records_format already defined by user - populate that
            kwargs['records_format'] = existing_records_format.\
                alter_hints({param_name: kwargs[param_name]})
        elif 'initial_hints' in self.possible_args:
            if 'initial_hints' not in kwargs:
                kwargs['initial_hints'] = {}
            kwargs['initial_hints'][param_name] = kwargs[param_name]
        elif 'records_format' in self.possible_args:
            # user must want a delimited records format - e.g., on
            # output where initial hints are not a thing as we're not
            # auto-detecting - but we haven't gotten their preferred
            # variant yet.  Let's assume the default variant to start -
            # maybe a future arg will reveal a more specific variant.
            kwargs['records_format'] =\
                DelimitedRecordsFormat(hints={param_name: kwargs[param_name]})  # type: ignore
        else:
            raise NotImplementedError(f"Could not find place for {param_name} in "
                                      f"{self.possible_args}")
        del kwargs[param_name]

    def fill_in_spectrum_base_url(self, kwargs: Dict[str, Any]) -> None:
        db_name = kwargs['db_name']
        db_facts = self.session.creds.db_facts(db_name)
        schema_name = kwargs['schema_name']
        key = f"redshift_spectrum_base_url_{schema_name}"
        if key not in db_facts:
            raise ValueError("Please specify spectrum_base_url")
        else:
            kwargs['spectrum_base_url'] = db_facts[key]  # type: ignore

    def fill_in_google_cloud_creds(self, kwargs: Dict[str, Any]) -> None:
        kwargs['google_cloud_creds'] =\
            self.session.creds.google_sheets(self.config['gcp_creds_name'])
        del kwargs['gcp_creds_name']

    def fill_in_existing_table_handling(self, kwargs: Dict[str, Any]) -> None:
        kwargs['existing_table_handling'] =\
            ExistingTableHandling[kwargs['existing_table'].upper()]
        del kwargs['existing_table']

    def fill_in_db_engine(self, kwargs: Dict[str, Any]) -> None:
        kwargs['db_engine'] = self.session.get_db_engine(kwargs['db_name'])

    def fill_in_format(self, kwargs: Dict[str, Any]) -> None:
        records_format: BaseRecordsFormat
        if kwargs['format'] == 'delimited':
            records_format = DelimitedRecordsFormat()
        elif kwargs['format'] == 'parquet':
            records_format = ParquetRecordsFormat()
        else:
            raise NotImplementedError(f"No such records format type: {kwargs['format']}")

        if 'records_format' in kwargs:
            existing_records_format = kwargs['records_format']
            if type(records_format) != type(existing_records_format):
                raise NotImplementedError('Hints are not compatible '
                                          'with records format '
                                          f'{existing_records_format.format_type}')
        else:
            kwargs['records_format'] = records_format

        del kwargs['format']

    def fill_in_variant(self, kwargs: Dict[str, Any]) -> None:
        if 'records_format' in kwargs:
            # We've already started filling this out with hints -
            # but now we know which variant they want
            kwargs['records_format'] = kwargs['records_format'].alter_variant(kwargs['variant'])
        else:
            hints: PartialRecordsHints = {}
            if 'initial_hints' in kwargs:
                # Given the user gave us a variant, we won't be
                # using "initial hints" to sniff with - the hint
                # parameters they gave us should be specified as
                # part of the records format instead.
                hints = kwargs['initial_hints']
                del kwargs['initial_hints']
            kwargs['records_format'] = DelimitedRecordsFormat(variant=kwargs['variant'],
                                                              hints=hints)
        del kwargs['variant']

    def to_args(self) -> Dict[str, Any]:
        kwargs = {}
        for name, value in self.config.items():
            kwargs[name] = value
        self.possible_args = self.all_arg_names(self.method)
        self.expected_args = self.non_default_arg_names(self.method)
        self.missing_args = self.expected_args - kwargs.keys()

        if 'spectrum_base_url' in self.possible_args and kwargs.get('spectrum_base_url') is None:
            self.fill_in_spectrum_base_url(kwargs)

        for arg in self.missing_args:
            if arg == 'self':
                continue
            elif arg == 'google_cloud_creds':
                self.fill_in_google_cloud_creds(kwargs)
            elif arg == 'db_engine':
                self.fill_in_db_engine(kwargs)
            else:
                raise NotImplementedError(f"Did not know how to handle {arg} - "
                                          f"got config {self.config}")
        if 'db_name' in kwargs:
            # Already used to feed different args above
            del kwargs['db_name']
        # for unit test coverage stability, this needs to be a list
        # comprehension (which maintains the order of the kwargs keys)
        # instead of subtracting possible_args from kwargs.keys(),
        # which becomes a set, whose order will vary when iterated
        # over from test run to test run.
        extra_args = [kwarg for kwarg in kwargs if kwarg not in self.possible_args]

        for arg in extra_args:
            if arg == 'existing_table':
                self.fill_in_existing_table_handling(kwargs)
            elif arg == 'variant':
                self.fill_in_variant(kwargs)
            elif arg == 'format':
                self.fill_in_format(kwargs)
            elif arg in HINT_NAMES:
                self.add_hint_parameter(kwargs, arg)
            else:
                raise NotImplementedError(f"Teach me how to pass in {arg}")
        return kwargs


def config_to_args(config: JobConfig,
                   method: Callable[..., Any],
                   session: Session) -> Dict[str, Any]:
    return ConfigToArgs(config, method, session).to_args()



class DeserializationError(Exception):
    """Raised for errors encountered while reading in JSON representations
    of a Records Schema
    """
    pass


class UnsupportedSchemaError(DeserializationError):
    """Raised if the version of the Records Schema read in isn't supported
    by this version of Records Mover.  This is usually a sign that you
    are using an older version of Records Mover to try to read a
    directory written by a newer version of Records Mover.
    """
    pass
